import datetime
import json
import os
import time

import numpy as np
import pytest
from PIL import Image

import weave
from weave.legacy.weave import context_state as _context
from weave.legacy.weave.ops_domain.run_history.context import (
    error_on_non_vectorized_history_transform,
)
from weave.legacy.weave.ops_domain.run_history.history_op_common import _without_tags
from weave.legacy.weave.wandb_client_api import wandb_gql_query
from weave.legacy.weave.wandb_interface import wandb_stream_table

HISTORY_OP_NAME = "history3"


def image():
    imarray = np.random.rand(100, 100, 3) * 255
    return Image.fromarray(imarray.astype("uint8")).convert("RGBA")


_loading_builtins_token = _context.set_loading_built_ins()


@weave.type()
class CustomHistoryTestType:
    a: int
    b: str


_context.clear_loading_built_ins(_loading_builtins_token)


def make_base_types():
    return {
        "number": 42,
        "string": "hello world",
        "boolean": True,
        "object": CustomHistoryTestType(1, "hi"),
        "custom": image(),
        "user_defined_timestamp": datetime.datetime.now(),
    }


def make_all_types():
    base_types = make_base_types()
    base_types["list"] = list(make_base_types().values()) + [
        {**make_base_types()},
        None,
    ]
    base_types["dict"] = {**make_base_types()}
    return base_types


rows_tests = [
    # Here we have 1 test per type for easy debugging
    *[[{k: v}] for k, v in make_all_types().items()],
    # # Here we have 1 test for all the types
    [make_all_types()],
    # Here is a nasty test with really hard unions
    [{"list_of": [1, 2, 3]}, {"list_of": [4, 5]}],
    [
        {
            "a": image(),
        },
        {
            "a": 1,
        },
    ],
    [{"list_of_image": [image()]}],
    [
        {"a": [{"b": [1]}, {"b": [2, 3]}]},
        {"a": [{"b": [4, 5, 6]}, {"b": [7, 8, 9, 10]}]},
    ],
    [
        {"a": 1, "b": "hi", "c": CustomHistoryTestType(2, "bye"), "i": image()},
        {"a": True, "b": True, "c": True, "i": True},
        {
            "a": [
                1,
                "hi",
                True,
                CustomHistoryTestType(2, "bye"),
                {"a": 1, "b": "hi", "c": CustomHistoryTestType(2, "bye"), "i": image()},
            ]
        },
        {
            "randomly": {
                "nested": {
                    "objects": {
                        "a": 1,
                        "b": "hi",
                        "c": CustomHistoryTestType(2, "bye"),
                        "i": image(),
                    }
                }
            }
        },
    ],
]


@pytest.mark.parametrize(
    "rows",
    rows_tests,
)
def test_end_to_end_stream_table_history_path_batch_1(user_by_api_key_in_env, rows):
    return do_test_end_to_end_stream_table_history_path(
        user_by_api_key_in_env.username, rows
    )


def do_test_end_to_end_stream_table_history_path(username, rows):
    def do_assertion(history_node, row_type, row_accumulator, user_logged_keys):
        row_object_type = make_optional_type(row_type.object_type)

        for key in user_logged_keys:
            column_node = history_node[key]
            assert_type_assignment(
                row_object_type.property_types[key], column_node.type.object_type
            )
            column_value = weave.use(column_node).to_pylist_tagged()
            expected = []

            for row in row_accumulator:
                expected.append(row.get(key))
            assert compare_objects(column_value, expected)

    do_batch_test(username, rows, do_assertion)


def test_nested_pick_via_dots(user_by_api_key_in_env):
    rows = [
        {
            "a": 1,
            "b": {
                "c": 2,
            },
            "d.e": 3,
        }
    ]

    def do_assertion(history_node, row_type, row_accumulator, user_logged_keys):
        with error_on_non_vectorized_history_transform():
            assert weave.use(history_node["a"]).to_pylist_tagged() == [1]
            assert weave.use(history_node["b"]).to_pylist_tagged() == [{"c": 2}]
            assert weave.use(history_node["b"]["c"]).to_pylist_tagged() == [2]
            assert weave.use(history_node["b.c"]).to_pylist_tagged() == [2]
            assert weave.use(history_node["d"]).to_pylist_tagged() == [{"e": 3}]
            assert weave.use(history_node["d.e"]).to_pylist_tagged() == [3]

    do_batch_test(user_by_api_key_in_env.username, rows, do_assertion)


def test_missing_data(user_by_api_key_in_env):
    rows = [{"a": "1", "b": "17", "c": "42"}, {"a": "2"}]

    def do_assertion(history_node, row_type, row_accumulator, user_logged_keys):
        nodes = [
            history_node[0]["a"],
            history_node[0]["b"],
            history_node[1]["a"],
            history_node[1]["b"],
        ]
        res = weave.use(nodes)
        assert res[0] == "1"
        assert res[1] == "17"
        assert res[2] == "2"
        assert res[3] == None

    do_batch_test(user_by_api_key_in_env.username, rows, do_assertion)


def do_batch_test(username, rows, do_assertion):
    row_accumulator, st, user_logged_keys = do_logging(username, rows)

    row_type = weave.types.TypeRegistry.type_of([{}, *row_accumulator])
    run_node = weave.legacy.weave.ops.project(st._entity_name, st._project_name).run(
        st._table_name
    )

    # First assertion is with liveset
    wait_for_x_times(
        lambda: history_is_uploaded(
            run_node,
            len(row_accumulator),
            len(row_type.object_type.property_types),
            st._entity_name,
            st._project_name,
            st._table_name,
        )
    )
    # Wait for files to be uploaded
    st._flush()
    wait_for_x_times(lambda: st._lite_run.pusher._incoming_queue.empty())
    wait_for_x_times(lambda: st._lite_run.pusher._event_queue.empty())
    wait_for_x_times(lambda: st._lite_run.stream._queue.empty())
    history_node = run_node._get_op(HISTORY_OP_NAME)()
    do_assertion(history_node, row_type, row_accumulator, user_logged_keys)
    st.finish()

    if os.environ.get("PARQUET_ENABLED"):
        # Second assertion is with parquet files
        wait_for_x_times(
            lambda: history_moved_to_parquet(
                st._entity_name,
                st._project_name,
                st._table_name,
            )
        )
        history_node = run_node._get_op(HISTORY_OP_NAME)()
        do_assertion(history_node, row_type, row_accumulator, user_logged_keys)


def do_logging(username, rows, finish=False):
    table_name = "test_table_" + str(int(time.time()))
    st = wandb_stream_table.StreamTable(
        table_name=table_name,
        project_name="dev_test_weave_ci",
        entity_name=username,
        _disable_async_file_stream=True,
    )

    row_accumulator = []

    all_keys = set(["_step"])
    for row in rows:
        st.log(row)
        new_row = {
            # Extra fields automatically added from the Run
            "_step": len(row_accumulator),
            "_timestamp": datetime.datetime.now().timestamp(),
            # Extra fields automatically added from the StreamTable
            "_client_id": "dummy",
            "timestamp": datetime.datetime.utcnow(),
            # User fields
            **row,
        }
        row_accumulator.append(new_row)
        all_keys.update(list(row.keys()))

    if finish:
        st.finish()

    return row_accumulator, st, all_keys


def history_is_uploaded(
    run_node, exp_len, exp_cols_len, entity_name, project_name, run_name
):
    history_node = run_node._get_op(HISTORY_OP_NAME)()
    run_data = get_raw_gorilla_history(
        entity_name,
        project_name,
        run_name,
    )
    history = run_data.get("parquetHistory", {})
    return (
        len(history.get("liveData", []))
        == exp_len
        == (run_data.get("historyKeys", {}).get("lastStep", -999) + 1)
        and history.get("parquetUrls") == []
        and exp_cols_len == len(history_node.type.object_type.property_types)
    )


def history_moved_to_parquet(entity_name, project_name, run_name):
    history = get_raw_gorilla_history(entity_name, project_name, run_name)
    return (
        history.get("parquetHistory", {}).get("liveData") == []
        and len(history.get("parquetHistory", {}).get("parquetUrls", [])) > 0
    )


def compare_objects(a, b):
    if isinstance(a, Image.Image) and isinstance(b, Image.Image):
        return a.tobytes() == b.tobytes()
    elif isinstance(a, list) and isinstance(b, list):
        return len(a) == len(b) and all(compare_objects(a_, b_) for a_, b_ in zip(a, b))
    elif isinstance(a, dict) and isinstance(b, dict):
        return len(a) == len(b) and all(
            k in b and compare_objects(a[k], b[k]) for k in a
        )
    elif isinstance(a, datetime.datetime) and isinstance(b, datetime.datetime):
        return int(a.timestamp() * 1000) == int(b.timestamp() * 1000)
    return a == b


def make_optional_type(type_: weave.types.Type):
    if isinstance(type_, weave.types.List):
        type_ = weave.types.List(make_optional_type(type_.object_type))
    elif isinstance(type_, weave.types.TypedDict):
        type_ = weave.types.TypedDict(
            {k: make_optional_type(v) for k, v in type_.property_types.items()}
        )
    return weave.types.optional(type_)


def assert_type_assignment(a, b):
    a = _without_tags(a)
    b = _without_tags(b)
    if weave.types.optional(weave.types.TypedDict({})).assign_type(
        a
    ) and weave.types.optional(weave.types.TypedDict({})).assign_type(b):
        for k, ptype in a.property_types.items():
            assert k in b.property_types
            assert_type_assignment(ptype, b.property_types[k])
        return
    assert a.assign_type(b)


def wait_for_x_times(for_fn, times=10, wait=1):
    done = False
    while times > 0 and not done:
        times -= 1
        done = for_fn()
        time.sleep(wait)
    assert times > 0


def get_raw_gorilla_history(entity_name, project_name, run_name):
    query = """query WeavePythonCG($entityName: String!, $projectName: String!, $runName: String! ) {
            project(name: $projectName, entityName: $entityName) {
                run(name: $runName) {
                    historyKeys
                    parquetHistory(liveKeys: ["_timestamp"]) {
                        liveData
                        parquetUrls
                    }
                }
            }
    }"""
    variables = {
        "entityName": entity_name,
        "projectName": project_name,
        "runName": run_name,
    }
    res = wandb_gql_query(query, variables)
    return res.get("project", {}).get("run", {})


@pytest.mark.skip(reason="Local Perf Testing")
@pytest.mark.parametrize(
    "n_rows, n_cols",
    [
        # (100, 100),
        (1500, 1500),
        (10_000, 1_000),
        (1_000_000, 1_000),
        (1_000_000, 10_000),
        (10_000_000, 10_000),
    ],
)
def test_stream_table_perf(user_by_api_key_in_env, n_rows, n_cols):
    print(f"Conducting StreamTable perf test using {n_rows} rows and {n_cols} cols")
    table_name = "test_table_" + str(int(time.time()))
    st = wandb_stream_table.StreamTable(
        table_name=table_name,
        project_name="dev_test_weave_ci",
        entity_name=user_by_api_key_in_env.username,
        _disable_async_file_stream=True,
    )
    timings = {
        "log": 0,
        "history2_refine": 0,
        "history2_fetch_100_cols": 0,
        "history3_refine": 0,
        "history3_fetch_100_cols": 0,
    }
    print_interval = int(n_rows / 10)
    timings["log"] = 0
    for i in range(n_rows):
        row = {}
        for j in range(n_cols):
            row[f"col_{j}"] = (i * j) + (i + j)
        timings["log"] -= time.time()
        st.log(row)
        timings["log"] += time.time()
        if i % print_interval == 0:
            print(f"Logged {i} rows")
        if i % 100 == 0:
            time.sleep(0.5)

    print("Finishing Run")
    timings["log"] -= time.time()
    st.finish()
    timings["log"] += time.time()
    print(f"Log Time: {timings['log']}")

    run_node = weave.legacy.weave.ops.project(st._entity_name, st._project_name).run(
        st._table_name
    )

    timings["history2_refine"] -= time.time()
    history2_node = run_node.history2()
    timings["history2_refine"] += time.time()
    print(f"History2 Refine Time: {timings['history2_refine']}")

    fetch_nodes = [history2_node[f"col_{i}"] for i in range(100)]
    timings["history2_fetch_100_cols"] -= time.time()
    res = weave.use(fetch_nodes)
    timings["history2_fetch_100_cols"] += time.time()
    print(f"History2 Fetch 100 Cols Time: {timings['history2_fetch_100_cols']}")

    timings["history3_refine"] -= time.time()
    history3_node = run_node.history3()
    timings["history3_refine"] += time.time()
    print(f"History Stream Refine Time: {timings['history3_refine']}")

    fetch_nodes = [history3_node[f"col_{i}"] for i in range(100)]
    timings["history3_fetch_100_cols"] -= time.time()
    res = weave.use(fetch_nodes)
    timings["history3_fetch_100_cols"] += time.time()
    print(f"History Stream Fetch 100 Cols Time: {timings['history3_fetch_100_cols']}")

    print(f"Perf for {n_rows} rows and {n_cols} cols:")
    print(json.dumps(timings, indent=2))
    assert timings == None



from weave.legacy.weave import api as weave
from weave.legacy.weave import ops, storage, weave_internal


def test_autocommit(cereal_csv):
    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "C"  # value before set is 'C'

    ops.set(csv[-1]["type"], "XXXX")

    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "XXXX"

    ops.set(csv[-1]["type"], "YY")

    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "YY"


def test_nonconst(cereal_csv):
    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "C"  # value before set is 'C'
    ops.set(csv[-1]["type"], "XXXX")
    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "XXXX"
    ops.set(csv[-1]["type"], "YY")
    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "YY"


def test_mutate_with_use(cereal_csv):
    csv = ops.local_path(cereal_csv).readcsv()
    ops.set(csv[-1]["type"], "XXXX")
    assert weave.use(csv[-1]["type"]) == "XXXX"
    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "XXXX"


def test_mutate_artifact():
    storage.save({"a": 5, "b": 6}, "my-dict:latest")
    dict_obj = ops.get("local-artifact:///my-dict:latest/obj")
    ops.set(dict_obj["a"], 17)
    assert weave.use(dict_obj["a"]) == 17


def test_csv_saveload_type(cereal_csv):
    csv = weave.use(ops.local_path(cereal_csv).readcsv())
    ref = storage.save(csv)
    storage.get(str(ref))


def test_skips_list_indexcheckpoint(cereal_csv):
    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "C"  # value before set is 'C'

    row = ops.List.__getitem__(ops.list_indexCheckpoint(csv), -1)
    ops.set(row["type"], "XXXX")

    csv = ops.local_path(cereal_csv).readcsv()
    assert weave.use(csv[-1]["type"]) == "XXXX"



from weave.legacy.weave import api as weave
from weave.legacy.weave import graph

from ...legacy.weave import node_ref


def test_node_to_ref():
    l = [{"a": 5, "b": 6}, {"a": 7, "b": 9}]
    l_node = weave.save(l, "my-l")
    node = l_node[0]["a"]
    ref = node_ref.node_to_ref(node)
    assert ref.name == "my-l"
    assert ref.extra == ["0", "a"]

    node2 = node_ref.ref_to_node(ref)
    assert (
        str(node2) == 'get("local-artifact:///my-l:fc48d97f6600ff9162ca/obj")[0]["a"]'
    )



import inspect
import typing

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import context_state, derive_op, op_def, registry_mem

# Contrary to the way it is read, the weave.class() decorator runs AFTER the
# inner methods are defined. Therefore, this function runs after the ops are
# defined/registered. The majority of the logic here is to update those ops (and
# any derived ops) with proper names and types.


def weave_class(weave_type: type[types.Type]):
    def wrap(target):
        # add self type to input_types if its not already defined.
        for _, member in inspect.getmembers(target):
            if isinstance(member, op_def.BoundOpDef):
                opdef = typing.cast(op_def.BoundOpDef, member)
                self_type = opdef.input_type.arg_types.get("self")
                if self_type is not None and self_type == types.UnknownType():
                    opdef.input_type.arg_types["self"] = weave_type()
                    # Now that we have a self_type, we may be able to derive
                    # ops.
                    derive_op.derive_ops(opdef)
                # Replace function op names with method op names
                current_name = opdef.name
                requires_rename = current_name.startswith("op-")
                if requires_rename:
                    new_name = "%s-%s" % (
                        weave_type.name,
                        current_name[3:],
                    )
                    loading_builtins = context_state._loading_built_ins.get()
                    if loading_builtins:
                        registry_mem.memory_registry.rename_op(
                            current_name,
                            new_name,
                        )

                        for derived_handler_id, op in member.derived_ops.items():
                            handler = derive_op.handler_for_id(derived_handler_id)
                            handler.handle_class_decorator_update(
                                op, weave_type, new_name
                            )
                    else:
                        opdef.op_def.name = new_name

        # Check __dict__ instead of using regular attribute access
        # because we want to add instance_classes even if it is already
        # set in a base class
        if weave_type.__dict__.get("instance_classes") is None:
            weave_type.instance_classes = target
            weave_type.instance_class = target
        return target

    return wrap



"""Helper methods for displaying pydantic models."""

from typing import Any

from pydantic import BaseModel
from rich.console import Console
from rich.table import Table

from weave.trace import util


def dict_to_table(d: dict[str, Any]) -> Table:
    """Create a two-column table from a dictionary."""
    table = Table(show_header=False)
    table.add_column("Key", justify="right", style="bold cyan")
    table.add_column("Value")
    for k, v in d.items():
        table.add_row(k, str(v))
    return table


def table_to_str(table: Table) -> str:
    """Render a Rich Table to a string."""
    use_emojis = not util.is_notebook()
    console = Console(emoji=use_emojis)
    with console.capture() as capture:
        console.print(table)
    return capture.get().strip()


def model_to_table(model: BaseModel) -> Table:
    """Create a two-column table from a Pydantic model."""
    return dict_to_table(model.model_dump())


def model_to_str(model: BaseModel) -> Table:
    """Create a string representation of a Pydantic model."""
    return table_to_str(model_to_table(model))



class Error(Exception):
    """Base class for exceptions in this module."""

    pass


class RequestTooLarge(Error):
    """Raised when a request is too large."""

    pass


class InvalidRequest(Error):
    """Raised when a request is invalid."""

    pass


class InsertTooLarge(Error):
    """Raised when a single insert is too large."""

    pass


class InvalidFieldError(Error):
    """Raised when a field is invalid."""

    pass



import dataclasses
import json
import textwrap
import typing

import numpy as np
import pyarrow as pa
import pyarrow.feather as pf
import pyarrow.parquet as pq
from pyarrow import compute as pc

py_type = type

from weave.legacy.weave import errors
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import artifact_fs, partial_object


def arrow_type_to_weave_type(pa_type: pa.DataType) -> types.Type:
    if pa_type == pa.null():
        return types.NoneType()
    elif pa.types.is_dictionary(pa_type):
        return arrow_type_to_weave_type(pa_type.value_type)
    elif pa_type == pa.string():
        return types.String()
    elif pa_type == pa.large_string():
        return types.String()
    elif pa_type == pa.int64() or pa_type == pa.int32():
        return types.Int()
    elif pa_type in [pa.float64(), pa.float32(), pa.float16()]:
        return types.Float()
    elif pa_type == pa.bool_():
        return types.Boolean()
    elif pa.types.is_temporal(pa_type):
        return types.Timestamp()
    elif pa.types.is_binary(pa_type):
        return types.Bytes()
    elif pa.types.is_list(pa_type):
        return types.List(arrow_field_to_weave_type(pa_type.value_field))
    elif pa.types.is_struct(pa_type):
        return types.TypedDict({f.name: arrow_field_to_weave_type(f) for f in pa_type})
    elif pa.types.is_union(pa_type):
        return types.UnionType(*[arrow_type_to_weave_type(f.type) for f in pa_type])
    raise errors.WeaveTypeError(
        "Type conversion not implemented for arrow type: %s" % pa_type
    )


def arrow_field_to_weave_type(pa_field: pa.Field) -> types.Type:
    t = arrow_type_to_weave_type(pa_field.type)
    if pa_field.nullable:
        return types.optional(t)
    return t


def arrow_schema_to_weave_type(schema) -> types.Type:
    return types.TypedDict({f.name: arrow_field_to_weave_type(f) for f in schema})


@dataclasses.dataclass(frozen=True)
class ArrowArrayType(types.Type):
    instance_classes = [pa.ChunkedArray, pa.ExtensionArray, pa.Array]
    name = "ArrowArray"

    object_type: types.Type = types.Any()
    UNION_PREFIX = "__weave_union"
    UNION_TO_STRUCT_TYPE_CODE_COLNAME = "__weave_type_code"
    EMPTY_STRUCT_DUMMY_FIELD_NAME = "__weave_empty_struct"

    @classmethod
    def type_of_instance(cls, obj: pa.Array):
        return cls(arrow_type_to_weave_type(obj.type))

    def save_instance(self, obj, artifact, name):
        table = pa.table({"arr": obj})
        with artifact.new_file(f"{name}.parquet", binary=True) as f:
            pq.write_table(table, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.parquet", binary=True) as f:
            deserialized = pq.read_table(f)
            deserialized = deserialized["arr"].combine_chunks()
            return deserialized


@dataclasses.dataclass(frozen=True)
class ArrowTableType(types.Type):
    instance_classes = pa.Table
    name = "ArrowTable"

    object_type: types.Type = types.Any()

    @classmethod
    def type_of_instance(cls, obj: pa.Table):
        return cls(arrow_schema_to_weave_type(obj.schema))

    def save_instance(self, obj, artifact, name):
        with artifact.new_file(f"{name}.parquet", binary=True) as f:
            pq.write_table(obj, f)

    def load_instance(self, artifact, name, extra=None):
        with artifact.open(f"{name}.parquet", binary=True) as f:
            return pq.read_table(f)


# This function is evil and breaks performance because of the dictionary_decode
# call. Don't use it. TODO: remove all calls
def arrow_as_array(obj) -> pa.Array:
    # assumes obj is table or array
    if isinstance(obj, pa.Table):
        return pa.StructArray.from_arrays(
            # TODO: we shouldn't need to combine chunks, we can produce this in the
            # original chunked form for zero copy
            [c.combine_chunks() for c in obj.columns],
            names=obj.column_names,
        )
    elif isinstance(obj, pa.ChunkedArray):
        return obj.combine_chunks()
    elif isinstance(obj, pa.DictionaryArray):
        return arrow_as_array(obj.dictionary_decode())
    elif not isinstance(obj, pa.Array):
        raise TypeError("Expected pyarrow Array or Table, got %s" % type(obj))
    return obj


def offsets_starting_at_zero(arr: pa.ListArray) -> pa.IntegerArray:
    """
    We often have code to operate on the elements of a list array. That code
    will do something like:
    ```
        offsets = arr.offsets
        elements = arr.flatten()
        transformed = do_something(elements)
        result = pa.ListArray.from_arrays(offsets, transformed)
    ```
    However, sometimes for performance reasons, the offsets do not start at 0.
    For example, when the ListArray is actually a slice of a larger underlying
    buffer. We should use this helper function anytime we are doing something
    like the above example.
    """
    raw_offsets = arr.offsets
    if len(raw_offsets) == 0:
        return raw_offsets
    first_value = raw_offsets[0]
    if first_value == 0:
        return raw_offsets
    else:
        return pc.subtract(raw_offsets, first_value)


@dataclasses.dataclass(frozen=True)
class ArrowWeaveListType(types.Type):
    _base_type = types.List
    name = "ArrowWeaveList"

    object_type: types.Type = types.Any()

    @classmethod
    def type_of_instance(cls, obj):
        return cls(obj.object_type)

    def save_instance(self, obj, artifact, name):
        # rewriting refs is disabled after a refactor. We need this to work
        # for media and caching with local artifacts to work. But we don't need
        # it for W&B production for weave0 use cases.
        # TODO: fix this

        # If we are saving to the same artifact as we were written to,
        # then we don't need to rewrite any references.

        # if obj._artifact == artifact:
        #     arrow_data = obj._arrow_data
        # else:
        #     # super().save_instance(obj, artifact, name)
        #     # return
        #     arrow_data = rewrite_weavelist_refs(
        #         obj._arrow_data, obj.object_type, obj._artifact, artifact
        #     )
        if not obj._artifact == artifact:
            # somehow we still need this block of code to get the test passing,
            # even though we don't use its result.
            rewrite_weavelist_refs(
                obj._arrow_data, obj.object_type, obj._artifact, artifact
            )

        # v1 AWL format
        # from weave.legacy.weave.arrow import convert

        # parquet_friendly = convert.to_parquet_friendly(obj)
        # table = pa.table({"arr": parquet_friendly._arrow_data})
        # with artifact.new_file(f"{name}.ArrowWeaveList.parquet", binary=True) as f:
        #     pq.write_table(table, f)

        # v2 AWL format
        # - if it's a top-level typeddict we don't save in nested sub-array.
        #   reading specific columns from a parquet or feather file only gives a perf
        #   improvement for top-level columns.
        # - saved as feather file instead of parquet. There are some issues in our
        #   to/from parquet conversion with unions. They can be triggered by using
        #   v1 format with the hypothesis tests.
        artifact.metadata["_weave_awl_format"] = 2
        if isinstance(obj.object_type, types.TypedDict):
            arrow_data = obj._arrow_data
            if not obj.object_type.property_types:
                arrow_data = pa.StructArray.from_arrays(
                    [pa.nulls(len(arrow_data))], names=["_dummy"]
                )
            table = pa.Table.from_arrays(
                [arrow_data.field(i) for i in range(len(arrow_data.type))],
                names=[f.name for f in arrow_data.type],
            )
        else:
            table = pa.table({"arr": obj._arrow_data})
        with artifact.new_file(f"{name}.ArrowWeaveList.feather", binary=True) as f:
            pf.write_feather(table, f)

        with artifact.new_file(f"{name}.ArrowWeaveList.type.json") as f:
            json.dump(obj.object_type.to_dict(), f)

    def load_instance(
        self, artifact: artifact_fs.FilesystemArtifact, name: str, extra=None
    ):
        with artifact.open(f"{name}.ArrowWeaveList.type.json") as f:
            object_type = json.load(f)
            object_type = types.TypeRegistry.type_from_dict(object_type)
        from weave.legacy.weave.arrow import list_

        if "_weave_awl_format" not in artifact.metadata:
            # v1 AWL format
            with artifact.open(f"{name}.ArrowWeaveList.parquet", binary=True) as f:
                table = pq.read_table(f)
            arr = table["arr"].combine_chunks()
            with list_.unsafe_awl_construction("load_from_parquet"):
                l = self.instance_class(arr, object_type=object_type, artifact=artifact)  # type: ignore
                from weave.legacy.weave.arrow import convert

                res = convert.from_parquet_friendly(l)
        elif artifact.metadata["_weave_awl_format"] == 2:
            # v2 AWL format
            with artifact.open(f"{name}.ArrowWeaveList.feather", binary=True) as f:
                table = pf.read_table(f)
            if isinstance(object_type, types.TypedDict):
                if not object_type.property_types:
                    arr = pa.repeat({}, len(table))
                else:
                    arr = pa.StructArray.from_arrays(
                        [table[i].combine_chunks() for i in range(len(table.schema))],
                        names=[f.name for f in table.schema],
                    )
            else:
                arr = table["arr"].combine_chunks()
            res = self.instance_class(arr, object_type=object_type, artifact=artifact)  # type: ignore
        else:
            raise ValueError(
                f"Unknown _weave_awl_format {artifact.metadata['_weave_awl_format']}"
            )

        res.validate()

        return res


def rewrite_weavelist_refs(arrow_data, object_type, source_artifact, target_artifact):
    if isinstance(object_type, partial_object.PartialObjectType):
        # PartialObject is a leaf type
        return arrow_data
    if _object_type_has_props(object_type):
        prop_types = _object_type_prop_types(object_type)

        # handle empty struct case - the case where the struct has no fields
        if len(prop_types) == 0:
            return arrow_data

        if isinstance(arrow_data, pa.Table):
            arrays = {}
            for col_name, col_type in prop_types.items():
                column = arrow_data[col_name]
                arrays[col_name] = rewrite_weavelist_refs(
                    column, col_type, source_artifact, target_artifact
                )
            return pa.table(arrays)
        elif isinstance(arrow_data, pa.ChunkedArray):
            arrays = {}
            unchunked = arrow_data.combine_chunks()
            for col_name, col_type in prop_types.items():
                column = unchunked.field(col_name)
                arrays[col_name] = rewrite_weavelist_refs(
                    column, col_type, source_artifact, target_artifact
                )
            return pa.StructArray.from_arrays(
                arrays.values(), names=arrays.keys(), mask=pa.compute.is_null(unchunked)
            )
        elif isinstance(arrow_data, pa.StructArray):
            arrays = {}
            for col_name, col_type in prop_types.items():
                column = arrow_data.field(col_name)
                arrays[col_name] = rewrite_weavelist_refs(
                    column, col_type, source_artifact, target_artifact
                )
            return pa.StructArray.from_arrays(
                arrays.values(),
                names=arrays.keys(),
                mask=pa.compute.is_null(arrow_data),
            )
        elif isinstance(arrow_data, pa.NullArray):
            return arrow_data
        else:
            raise errors.WeaveTypeError('Unhandled type "%s"' % type(arrow_data))
    elif isinstance(object_type, types.UnionType):
        non_none_members = [
            m for m in object_type.members if not isinstance(m, types.NoneType)
        ]
        nullable = len(non_none_members) < len(object_type.members)
        if len(non_none_members) > 1:
            arrow_data = arrow_as_array(arrow_data)
            if not isinstance(arrow_data.type, pa.UnionType):
                raise errors.WeaveTypeError(
                    "Expected UnionType, got %s" % type(arrow_data.type)
                )
            arrays = []
            for i, _ in enumerate(arrow_data.type):
                rewritten = rewrite_weavelist_refs(
                    arrow_data.field(i),
                    non_none_members[i]
                    if not nullable
                    else types.UnionType(types.NoneType(), non_none_members[i]),
                    source_artifact,
                    target_artifact,
                )
                arrays.append(rewritten)
            return pa.UnionArray.from_dense(
                arrow_data.type_codes, arrow_data.offsets, arrays
            )
        return rewrite_weavelist_refs(
            arrow_data, types.non_none(object_type), source_artifact, target_artifact
        )
    elif _object_type_is_basic(object_type):
        return arrow_data
    elif isinstance(object_type, (types.List, ArrowWeaveListType)):
        data = arrow_data
        if isinstance(arrow_data, pa.ChunkedArray):
            data = arrow_data.combine_chunks()
        return pa.ListArray.from_arrays(
            offsets_starting_at_zero(data),
            rewrite_weavelist_refs(
                data.flatten(),
                object_type.object_type,
                source_artifact,
                target_artifact,
            ),
            mask=pa.compute.is_null(data),
        )
    else:
        # We have a column of refs
        new_refs = []
        for ref_str in arrow_data:
            ref_str = ref_str.as_py()
            new_refs.append(
                _rewrite_ref_for_save(
                    ref_str, object_type, source_artifact, target_artifact
                )
                if ref_str is not None
                else None
            )
        return pa.array(new_refs)


def _object_type_has_props(object_type):
    from weave.legacy.weave.language_features.tagging import tagged_value_type

    return (
        isinstance(object_type, types.TypedDict)
        or isinstance(object_type, types.ObjectType)
        or isinstance(object_type, tagged_value_type.TaggedValueType)
    )


def _object_type_prop_types(object_type):
    from weave.legacy.weave.language_features.tagging import tagged_value_type

    if isinstance(object_type, tagged_value_type.TaggedValueType):
        return {
            "_tag": object_type.tag,
            "_value": object_type.value,
        }
    prop_types = object_type.property_types
    if callable(prop_types):
        prop_types = prop_types()
    return prop_types


def _object_type_is_basic(object_type):
    if isinstance(object_type, types.Const):
        object_type = object_type.val_type
    return isinstance(object_type, types.BasicType) or (
        isinstance(object_type, (types.Timestamp, types.TimeDelta))
    )


def _rewrite_ref_for_save(entry: str, object_type, source_artifact, target_artifact):
    # TODO: This should be simpler. We take a ref string, which may be absolute or
    # relative, and want to ensure the object is referencable from the target_artifact.
    # This function should be written in those terms.
    if ":" in entry:
        # Already a URI
        return entry

    if isinstance(source_artifact, artifact_fs.FilesystemArtifact):
        # Our source is an artifact, construct a full URI.
        return str(source_artifact.ref_from_local_str(entry, object_type).uri)

    # Source is ObjLookupMem, save to the artifact
    return target_artifact.set(
        entry, object_type, source_artifact.get(entry, object_type)
    ).local_ref_str()


def pretty_print_arrow_type(t: typing.Union[pa.Schema, pa.DataType, pa.Field]) -> str:
    if isinstance(t, pa.Schema):
        return "Schema:\n" + textwrap.indent(
            "\n".join(pretty_print_arrow_type(f) for f in t), "  "
        )
    elif isinstance(t, pa.Field):
        return f"{t.name}: {'nullable' if t.nullable else 'non-null'} {pretty_print_arrow_type(t.type)}"

    if isinstance(t, pa.StructType):
        return "Struct:\n" + textwrap.indent(
            "\n".join(pretty_print_arrow_type(f) for f in t), "  "
        )

    elif isinstance(t, pa.ListType):
        return "List\n" + textwrap.indent(pretty_print_arrow_type(t.value_type), "  ")

    elif isinstance(t, pa.UnionType):
        return "Union:\n" + textwrap.indent(
            "\n".join(pretty_print_arrow_type(f) for f in t), "  "
        )

    return str(t)


def union_is_null(arr: pa.Array):
    # pyarrow doesn't support is_null on unions. We do it ourselves
    # by checking each child entry for whether its null. This code
    # is vectorized so still fast.
    merged = pa.nulls(len(arr), pa.bool_())
    for type_code in range(len(arr.type)):
        field = pa.compute.is_null(arr.field(type_code))
        if len(field) == 0:
            continue
        mask = pa.compute.equal(arr.type_codes, type_code)
        indexes = pa.compute.multiply(mask.cast(pa.int8()), arr.offsets)
        values = field.take(indexes)
        merged = pa.compute.if_else(mask, values, merged)
    return merged


def safe_is_null(arr: pa.Array):
    # is_null that also handles unions correctly since pyarrow doesn't.
    if isinstance(arr.type, pa.UnionType):
        return union_is_null(arr)
    else:
        return pa.compute.is_null(arr)


def safe_coalesce(*arrs: pa.Array):
    if not arrs:
        raise ValueError("coalesce requires at least one argument")
    # pyarrows coalesce doesn't handle unions. This does!
    if len(arrs[0]) == 0:
        return arrs[0]
    result = arrs[0]
    for arr in arrs[1:]:
        result = pa.compute.if_else(safe_is_null(result), arr, result)
    return result


def arrow_zip(*arrs: pa.Array) -> pa.Array:
    n_arrs = len(arrs)
    output_len = min(len(a) for a in arrs)
    array_indexes = np.tile(np.arange(n_arrs, dtype="int64"), output_len)
    item_indexes = np.floor(np.arange(0, output_len, 1.0 / n_arrs)).astype("int64")
    indexes = item_indexes + array_indexes * output_len
    concatted = pa.concat_arrays(arrs)
    interleaved = concatted.take(indexes)
    offsets = np.arange(0, len(interleaved) + len(arrs), len(arrs), dtype="int64")
    return pa.ListArray.from_arrays(offsets, interleaved)



import pytest

import weave
from weave.legacy.weave import artifact_fs, artifact_local, storage
from weave.legacy.weave import ops_arrow as arrow


def test_artifact():
    data = arrow.to_arrow([{"a": 5}])
    data_ref = storage.save(data)
    assert data_ref.type == arrow.ArrowWeaveListType(
        object_type=weave.types.TypedDict(property_types={"a": weave.types.Int()})
    )
    assert data_ref.get() == data

    art = data_ref.artifact
    assert weave.type_of(art) == artifact_local.LocalArtifactType()
    art_ref = storage.save(art)
    # in the special case of saving an artifact, we have a ref to that artifact
    # instead of a new artifact
    assert art_ref.artifact == art and art_ref.path is None
    assert art_ref.get() == art
    assert art_ref.type == artifact_local.LocalArtifactType()

    art_dir = art.path_info("")
    assert weave.type_of(art_dir) == artifact_fs.FilesystemArtifactDirType()
    art_dir_ref = storage.save(art_dir)
    assert art_dir_ref.get() == art_dir
    assert art_dir_ref.type == artifact_fs.FilesystemArtifactDirType()

    art_dir_files = art_dir.files
    exp_art_file1_type = artifact_fs.FilesystemArtifactFileType(
        extension=weave.types.Const(weave.types.String(), "feather"),
        wbObjectType=weave.types.NoneType(),
    )
    assert weave.type_of(art_dir_files) == weave.types.TypedDict(
        {
            "obj.ArrowWeaveList.feather": artifact_fs.FilesystemArtifactFileType(
                extension=weave.types.Const(weave.types.String(), "feather"),
                wbObjectType=weave.types.NoneType(),
            ),
            "obj.ArrowWeaveList.type.json": artifact_fs.FilesystemArtifactFileType(
                extension=weave.types.Const(weave.types.String(), "json"),
                wbObjectType=weave.types.TypeType(attr_types={}),
            ),
            "obj.type.json": artifact_fs.FilesystemArtifactFileType(
                extension=weave.types.Const(weave.types.String(), "json"),
                wbObjectType=weave.types.TypeType(attr_types={}),
            ),
        }
    )
    assert len(art_dir.files) == 3

    art_file1 = art_dir_files["obj.ArrowWeaveList.feather"]
    assert art_file1 == artifact_fs.FilesystemArtifactFile(
        art, "obj.ArrowWeaveList.feather"
    )

    # Here we don't create a new object
    art_file1_ref = storage.save(art_file1)
    assert (
        art_file1_ref.artifact == art
        and art_file1_ref.path == "obj.ArrowWeaveList.feather"
    )
    assert art_file1_ref.type == exp_art_file1_type

    assert len(art_dir.dirs) == 0


def test_local_artifact_name():
    with pytest.raises(ValueError):
        artifact_local.LocalArtifact("a/b")
    with pytest.raises(ValueError):
        artifact_local.LocalArtifact("a\\b")
    with pytest.raises(ValueError):
        artifact_local.LocalArtifact("a:b")
    with pytest.raises(ValueError):
        artifact_local.LocalArtifact("a..b")



import ast
import builtins
import collections
import inspect
import json
import os
import sys
import textwrap
import types as py_types
import typing
from _ast import AsyncFunctionDef, ExceptHandler
from typing import Any

from weave.legacy.weave import storage
from weave.legacy.weave import weave_types as types
from weave.legacy.weave import (
    artifact_fs,
    artifact_local,
    context_state,
    environment,
    errors,
    infer_types,
    registry_mem,
)

if typing.TYPE_CHECKING:
    from .op_def import OpDef


def type_code(type_):
    if isinstance(type_, py_types.GenericAlias) or isinstance(
        type_, typing._GenericAlias  # type: ignore
    ):
        args = ", ".join(type_code(t) for t in type_.__args__)
        if type_.__origin__ == list or type_.__origin__ == collections.abc.Sequence:
            return f"list[{args}]"
        elif type_.__origin__ == dict:
            return f"dict[{args}]"
        elif type_.__origin__ == typing.Union:
            return f"typing.Union[{args}]"
        else:
            return f"{type_.__origin__}[{args}]"
    else:
        return type_.__name__


def generate_referenced_type_code(type_):
    # Given a function that may have type annotations, generate non-redundant
    # code that declares any referenced types and their referenced types and
    # so on.

    # Absolutely horrible and hacky. This is not recursive in a tree/graph
    # way, its linear, so it'll only produce one TypedDict if there are many.
    # Using this to get the versioned object notebook working.
    if infer_types.is_typed_dict_like(type_):
        result = f"class {type_.__name__}(typing.TypedDict):\n"
        for k in type_.__annotations__:
            result += f"    {k}: {type_code(type_.__annotations__[k])}\n"
        return result
    elif isinstance(type_, py_types.GenericAlias) or isinstance(
        type_, typing._GenericAlias  # type: ignore
    ):
        return generate_referenced_type_code(type_.__args__[0])


def arg_names(args: ast.arguments):
    arg_names = set()
    for arg in args.args:
        arg_names.add(arg.arg)
    for arg in args.kwonlyargs:
        arg_names.add(arg.arg)
    for arg in args.posonlyargs:
        arg_names.add(arg.arg)
    if args.vararg:
        arg_names.add(args.vararg.arg)
    if args.kwarg:
        arg_names.add(args.kwarg.arg)
    return arg_names


class ExternalVariableFinder(ast.NodeVisitor):
    """Given a an AST of a python function, find all external variable references.

    External variable references are variables that are used in the function but
    not created in the function.

    The general strategy is to walk the AST keeping track of what variables are in
    scope. Any variables loads we encounter that are not in the current scope are
    external.
    """

    def __init__(self):
        self.external_vars = {}
        self.scope_stack = [set()]  # Start with a global scope

    def visit_Import(self, node):
        for alias in node.names:
            self.scope_stack[-1].add(alias.asname or alias.name)

    def visit_ImportFrom(self, node):
        for alias in node.names:
            self.scope_stack[-1].add(alias.asname or alias.name)

    def visit_FunctionDef(self, node):
        # Add function name to the current scope
        self.scope_stack[-1].add(node.name)
        # Add function arguments to new scope
        self.scope_stack.append(arg_names(node.args))
        self.generic_visit(node)
        self.scope_stack.pop()  # Pop function scope when we exit the function

    def visit_AsyncFunctionDef(self, node) -> Any:
        self.scope_stack[-1].add(node.name)
        self.scope_stack.append(arg_names(node.args))
        self.generic_visit(node)
        self.scope_stack.pop()  # Pop function scope when we exit the function

    def visit_Lambda(self, node):
        # Add function arguments to the current scope
        self.scope_stack.append(arg_names(node.args))
        self.generic_visit(node)
        self.scope_stack.pop()  # Pop function scope when we exit the function

    def visit_ListComp(self, node) -> Any:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.elt)

    def visit_SetComp(self, node) -> Any:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.elt)

    def visit_GeneratorExp(self, node) -> Any:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.elt)

    def visit_DictComp(self, node) -> Any:
        # Change visit order, visit generators first which is where variable
        # definitions happen
        for generator in node.generators:
            self.visit(generator)
        self.visit(node.key)
        self.visit(node.value)

    def visit_ExceptHandler(self, node: ExceptHandler) -> Any:
        self.scope_stack[-1].add(node.name)
        self.generic_visit(node)

    def visit_Name(self, node):
        # print("  VISIT NAME", node.id, node.ctx)
        # If a variable is used (loaded) but not defined in any scope in the stack, and not builtin it's external
        # TODO: we don't capture python version, but builtins can change from version to version!
        if isinstance(node.ctx, ast.Store):
            self.scope_stack[-1].add(node.id)
        elif isinstance(node.ctx, ast.Load) and not any(
            node.id in scope for scope in self.scope_stack
        ):
            self.external_vars[node.id] = True


def resolve_var(fn: typing.Callable, var_name: str):
    """Given a python function, resolve a non-local variable name."""
    # First to see if the variable is in the closure
    if fn.__closure__:
        closure_vars = {}
        # __code__.co_freevars is the closure variable names in order
        for vn, closure_cell in zip(fn.__code__.co_freevars, fn.__closure__):
            closure_vars[vn] = closure_cell.cell_contents
        if var_name in closure_vars:
            return closure_vars[var_name]
    if var_name in fn.__globals__:
        return fn.__globals__[var_name]
    return None


class RefJSONEncoder(json.JSONEncoder):
    """Json encoder used for convert storage.to_json_with_refs result to python code"""

    SPECIAL_REF_TOKEN = "__WEAVE_REF__"

    def default(self, o):
        from .weave_client import ObjectRef

        ref_code = None
        if isinstance(o, artifact_fs.FilesystemArtifactRef):
            if o.serialize_as_path_ref:
                ref_code = f"weave.storage.artifact_path_ref('{o.local_ref_str()}')"
            else:
                ref_code = f"weave.ref('{str(o)}')"
        elif isinstance(o, (ObjectRef)):
            ref_code = f"weave.ref('{str(o)}')"

        if ref_code is not None:
            # This will be a quoted json string in the json.dumps result. We put special
            # tokens in so we can remove the quotes in the final result
            return f"{self.SPECIAL_REF_TOKEN}{ref_code}.get(){self.SPECIAL_REF_TOKEN}"
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, o)


class GetCodeDepsResult(typing.TypedDict):
    import_code: list[str]
    code: list[str]
    warnings: list[str]


def get_code_deps(
    fn: typing.Callable, artifact: artifact_fs.FilesystemArtifact
) -> GetCodeDepsResult:
    """Given a python function, return source code that contains the dependencies of that function.

    This will:
    - include any functions within the same module that are referenced within the function body
    - import any modules that are referenced within the function body or any referenced function bodies

    Current issues (to fix):
    - imported modules may be within the function's package, which doesn't work (any imported modules need
      to be present in the loading code)
    - doesn't serialize python requirements and other necessary system information
    - type annotations are not well handled, and may cause errors
    - refering to @weave.type() objects will cause errors

    Args:
        fn: The Python function to analyze.

    Returns:
        tupe: A tuple containing
            import_code: str, the code that should be included in the generated code to ensure all
                dependencies are available for the function body.
            warnings: list[str], any warnings that occurred during the process.
    """
    print("USING OP DEF TYPE VERSION")
    # Generates repeats.
    warnings: list[str] = []

    source = textwrap.dedent(inspect.getsource(fn))
    try:
        parsed = ast.parse(source)
    except SyntaxError:
        warnings.append(f"Could not parse source of function {fn}.")
        return {"import_code": [], "code": [], "warnings": warnings}

    visitor = ExternalVariableFinder()
    visitor.visit(parsed)
    external_vars = list(visitor.external_vars)

    import_code = []
    code = []
    for var_name in external_vars:
        var_value = resolve_var(fn, var_name)
        # var_value = fn.__globals__.get(var_name)
        if var_value is None:
            if getattr(builtins, var_name, None):
                # Its a builtin, carry on
                continue
            warnings.append(
                f'Could not resolve var "{var_name}" declared in body of fn {fn}. This op will not be reloadable, but calls to it will be tracked'
            )
        elif isinstance(var_value, py_types.ModuleType):
            import_line = f"import {var_value.__name__}"
            if var_value.__name__ != var_name:
                import_line += f" as {var_name}"
            import_code.append(import_line)
        elif isinstance(var_value, py_types.FunctionType):
            if var_value.__module__ == fn.__module__:
                # For now, if the function is in another module.
                # we just import it. This is ok for libraries, but not
                # if the user has functions declared within their
                # package but outside of the op module.
                result = get_code_deps(var_value, artifact)
                fn_warnings = result["warnings"]
                fn_import_code = result["import_code"]
                fn_code = result["code"]

                import_code += fn_import_code

                code += fn_code
                code.append(inspect.getsource(var_value))

                warnings += fn_warnings
            else:
                if var_value.__module__.split(".")[0] == fn.__module__.split(".")[0]:
                    pass

                import_line = f"from {var_value.__module__} import {var_value.__name__}"
                if var_value.__name__ != var_name:
                    import_line += f"as {var_name}"

                import_code.append(import_line)

        else:
            if (
                hasattr(var_value, "__name__")
                and hasattr(var_value, "__module__")
                and var_value.__module__ != fn.__module__
            ):
                import_line = f"from {var_value.__module__} import {var_value.__name__}"
                if var_value.__name__ != var_name:
                    import_line += f"as {var_name}"
                import_code.append(import_line)
            else:
                try:
                    json_val = storage.to_json_with_refs(
                        var_value, artifact, path=[var_name]
                    )
                except (errors.WeaveTypeError, errors.WeaveSerializeError) as e:
                    warnings.append(
                        f"Serialization error for value of {var_name} needed by {fn}. Encountered:\n    {e}"
                    )
                else:
                    code_paragraph = (
                        f"{var_name} = "
                        + json.dumps(json_val, cls=RefJSONEncoder, indent=4)
                        + "\n"
                    )
                    code_paragraph = code_paragraph.replace(
                        f'"{RefJSONEncoder.SPECIAL_REF_TOKEN}', ""
                    )
                    code_paragraph = code_paragraph.replace(
                        f'{RefJSONEncoder.SPECIAL_REF_TOKEN}"', ""
                    )
                    code.append(code_paragraph)
    return {"import_code": import_code, "code": code, "warnings": warnings}


def get_import_statements_for_annotations(func) -> list[str]:
    """Ensures we have imports for all the types used in the function annotations."""
    annotations = func.__annotations__
    imports_needed = set()

    for annotation in annotations.values():
        if hasattr(annotation, "__module__") and hasattr(annotation, "__name__"):
            module_name = annotation.__module__
            class_name = annotation.__name__
            if module_name != "builtins":
                imports_needed.add(f"from {module_name} import {class_name}")

    return list(imports_needed)


def find_last_weave_op_function(source_code: str):
    """Given a string of python source code, find the last function that is decorated with 'weave.op'."""
    tree = ast.parse(source_code)

    last_function = None

    for node in ast.walk(tree):
        if isinstance(node, ast.FunctionDef) or isinstance(node, AsyncFunctionDef):
            for decorator in node.decorator_list:
                # Check if the decorator is 'weave.op'
                if isinstance(decorator, ast.Name) and decorator.id == "weave.op":
                    last_function = node
                    break  # Break the inner loop, continue with the next function
                # Check if the decorator is a call to 'weave.op()'
                elif (
                    isinstance(decorator, ast.Call)
                    and isinstance(decorator.func, ast.Attribute)
                    and decorator.func.attr == "op"
                    and isinstance(decorator.func.value, ast.Name)
                    and decorator.func.value.id == "weave"
                ):
                    last_function = node
                    break  # Break the inner loop, continue with the next function

    return last_function


def dedupe_list(original_list: list[str]) -> list[str]:
    seen = set()
    deduped = []
    for x in original_list:
        if x not in seen:
            deduped.append(x)
            seen.add(x)
    return deduped


class OpDefType(types.Type):
    def save_instance(self, obj: "OpDef", artifact, name):
        if obj.name.startswith("mapped_"):
            # Skip mapped (derived ops)
            return None
        if obj.is_builtin:
            with artifact.new_file(f"{name}.json") as f:
                json.dump({"name": obj.name}, f)
        else:
            # Type annotations are not well handled at the moment.
            # Get import statements for any annotations
            # code += (
            #     "\n".join(get_import_statements_for_annotations(obj.raw_resolve_fn))
            #     + "\n\n"
            # )

            result = get_code_deps(obj.raw_resolve_fn, artifact)
            import_code = result["import_code"]
            code = result["code"]
            warnings = result["warnings"]
            if warnings:
                message = f"Warning: Incomplete serialization for op {obj}. This op may not be reloadable"
                for warning in warnings:
                    message += "\n  " + warning
                if context_state.get_legacy_strict_op_saving():
                    raise errors.WeaveOpSerializeError(message)
                else:
                    # print(message)
                    pass

            # This is hacky handling of TypedDict annotations. Fixes
            # 2_images_gen notebook test.
            # Create TypedDict types for referenced TypedDicts
            resolve_annotations = obj.raw_resolve_fn.__annotations__
            for k, type_ in resolve_annotations.items():
                gen_type_code = generate_referenced_type_code(type_)
                if gen_type_code is not None:
                    code.append(gen_type_code)
                    if "typing" in gen_type_code:
                        import_code.insert(0, "import typing")

            code.append(textwrap.dedent(inspect.getsource(obj.raw_resolve_fn)))
            with artifact.new_file(f"{name}.py") as f:
                import_block = "\n".join(import_code)
                import_lines = ["import weave"] + import_block.split("\n")
                import_lines = dedupe_list(import_lines)
                import_lines = [l for l in import_lines if "weave.api" not in l]
                import_block = "\n".join(import_lines)
                code_block = "\n".join(code)
                f.write(f"{import_block}\n\n{code_block}")

    def load_instance(cls, artifact, name, extra=None):
        if environment.wandb_production():
            # Returning None here instead of erroring allows the Weaveflow app
            # to reference op defs without crashing.
            return None
        try:
            with artifact.open(f"{name}.json") as f:
                op_spec = json.load(f)

            return registry_mem.memory_registry._ops[op_spec["name"]]
        except FileNotFoundError:
            pass

        from weave.legacy.weave import artifact_wandb

        is_wandb_artifact = False
        if isinstance(artifact, artifact_wandb.WandbArtifact):
            is_wandb_artifact = True

        file_name = f"{name}.py"
        module_path = artifact.path(file_name)

        if is_wandb_artifact:
            module_dir = os.path.dirname(module_path)
            import_name = os.path.splitext(os.path.basename(module_path))[0]
        else:
            # Python import caching means we can't just import "obj.py" here
            # because serialized ops would be cached at the "obj" key. We include
            # the version in the module name to avoid this. Since version names
            # are content hashes, this is correct.
            #
            art_and_version_dir = module_path[: -(1 + len(file_name))]
            art_dir, version_subdir = art_and_version_dir.rsplit("/", 1)
            module_dir = art_dir
            import_name = (
                version_subdir
                + "."
                + ".".join(os.path.splitext(file_name)[0].split("/"))
            )

        sys.path.insert(0, os.path.abspath(module_dir))
        with context_state.no_op_register():
            try:
                mod = __import__(import_name, fromlist=[module_dir])
            except Exception as e:
                print("Op loading exception. This might be fine!", e)
                import traceback

                traceback.print_exc()
                return None
        sys.path.pop(0)

        # In the case where the saved op calls another op, we will have multiple
        # ops in the file. The file will look like
        # op1 = weave.ref('...').get()
        # @weave.op()
        # def op2():
        #     op1()
        #
        # We need to get the last declared op in the module. We can't do this
        # simply by iterating module attributes, because order is not guaranteed,
        # so we resort to looking at the source ast.
        last_op_function = find_last_weave_op_function(inspect.getsource(mod))
        if last_op_function is None:
            print(
                f"Unexpected Weave module saved in: {module_path}. No op defs found. All members: {dir(mod)}. {module_dir=} {import_name=}"
            )
            return None

        od: "OpDef" = getattr(mod, last_op_function.name)

        try:
            location = artifact.uri_obj.with_path(name)
            registry_mem.memory_registry.register_op(od, location=location)
        except NotImplementedError:
            pass

        return od


def fully_qualified_opname(wrap_fn):
    op_module_file = os.path.abspath(inspect.getfile(wrap_fn))
    if op_module_file.endswith(".py"):
        op_module_file = op_module_file[:-3]
    elif op_module_file.endswith(".pyc"):
        op_module_file = op_module_file[:-4]
    return "file://" + op_module_file + "." + wrap_fn.__name__



"""Support for a collection of refs."""

from typing import Iterable, Optional

from rich.table import Table

from weave.trace.client_context import weave_client as weave_client_context
from weave.trace.refs import AnyRef, CallRef, parse_uri
from weave.trace.rich_container import AbstractRichContainer
from weave.trace.vals import WeaveObject


class Refs(AbstractRichContainer[str]):
    """A collection of ref strings with utilities."""

    def __init__(self, refs: Optional[Iterable[str]] = None) -> None:
        super().__init__("Ref", refs)

    def _add_table_columns(self, table: Table) -> None:
        table.add_column("Ref", overflow="fold")

    def _item_to_row(self, item: str) -> list:
        return [item]

    def parsed(self) -> list[AnyRef]:
        return [parse_uri(ref) for ref in self.items]

    def call_refs(self) -> "Refs":
        return Refs(ref for ref in self.items if isinstance(parse_uri(ref), CallRef))

    # TODO: Perhaps there should be a Calls that extends AbstractRichContainer
    def calls(self) -> list[WeaveObject]:
        client = weave_client_context.require_weave_client()
        objs = []
        for ref in self.call_refs():
            parsed = parse_uri(ref)
            assert isinstance(parsed, CallRef)
            objs.append(client.get_call(parsed.id))
        return objs



import weave
from weave.legacy.weave import ops_arrow


def test_with_columns_basic():
    t = weave.save(
        ops_arrow.to_arrow([{"a": 5, "b": 6, "d": {"y": 9}}, {"a": 7, "d": {"y": 11}}])
    )
    result_node = t.with_columns(
        {
            # should insert entries into d
            "d.x": ops_arrow.to_arrow([1, 2]),
            # should overwrite existing a
            "a": ops_arrow.to_arrow([-1, -2]),
        }
    )
    result = weave.use(result_node)
    assert result_node.type == ops_arrow.ArrowWeaveListType(
        weave.types.TypedDict(
            {
                "b": weave.types.optional(weave.types.Int()),
                "d": weave.types.TypedDict(
                    {"y": weave.types.Int(), "x": weave.types.Int()}
                ),
                "a": weave.types.Int(),
            }
        )
    )
    assert result.to_pylist_tagged() == [
        {"b": 6, "d": {"y": 9, "x": 1}, "a": -1},
        {"b": None, "d": {"y": 11, "x": 2}, "a": -2},
    ]


def test_with_columns_non_dict():
    t = weave.save(
        ops_arrow.to_arrow([{"a": 5, "b": 6, "d": {"y": 9}}, {"a": 7, "d": {"y": 11}}])
    )
    result_node = t.with_columns(
        {
            # a should be replaced with dict
            "a.x": ops_arrow.to_arrow([1, 2]),
        }
    )
    result = weave.use(result_node)
    assert result_node.type == ops_arrow.ArrowWeaveListType(
        weave.types.TypedDict(
            {
                "b": weave.types.optional(weave.types.Int()),
                "d": weave.types.TypedDict({"y": weave.types.Int()}),
                "a": weave.types.TypedDict({"x": weave.types.Int()}),
            }
        )
    )
    assert result.to_pylist_tagged() == [
        {"b": 6, "d": {"y": 9}, "a": {"x": 1}},
        {"b": None, "d": {"y": 11}, "a": {"x": 2}},
    ]


def test_with_columns_length_mismatch():
    t = weave.save(
        ops_arrow.to_arrow([{"a": 5, "b": 6, "d": {"y": 9}}, {"a": 7, "d": {"y": 11}}])
    )
    result = weave.use(t.with_columns({"a": ops_arrow.to_arrow([1])}))
    assert result == None
    result = weave.use(t.with_columns({"a": ops_arrow.to_arrow([1, 2, 3])}))
    assert result == None



# Top-Level Namespace: weave.monitoring

# TODO: Notebook Walkthrough

from weave.legacy.weave.monitoring.monitor import (
    default_monitor,
    deinit_monitor,
    init_monitor,
)
from weave.legacy.weave.wandb_interface.wandb_stream_table import StreamTable

__all__ = ["StreamTable", "monitor", "openai"]



import pytest

from weave.legacy.weave import uris
from weave.legacy.weave.artifact_wandb import WeaveWBArtifactURI

URIS = [
    "op-get",
    "op-get/version1",
    "local-artifact:///op-test:version",
    "local-artifact:///op-test:version#1",
    "local-artifact:///op-test:version#1/2/3",
    "local-artifact:///op-test:version/a.txt#1",
    "local-artifact:///op-test:version/a.txt",
    "local-artifact:///op-test:version/subdir/a.txt",
    "wandb-artifact:///entity-name/project/op-test:v0",
    "wandb-artifact:///entity-name/project/op-test:v0#1",
    "wandb-artifact:///entity-name/project/op-test:v0#1/2/3",
    "wandb-artifact:///entity-name/project/op-test:v0/obj.types.json#1",
]


@pytest.mark.parametrize("uri_str", URIS)
def test_parse_uri_str(uri_str: str):
    local_uri = uris.WeaveURI.parse(uri_str)
    assert uri_str == str(local_uri)


def test_uri_with_slashes():
    uri = WeaveWBArtifactURI(
        "test-art-deleted-07/02/23-17898374309789951906",
        "latest",
        "test",
        "test",
        "",
        "test.table.json",
    )
    uri_str = str(uri)
    uri2 = uris.WeaveURI.parse(uri_str)
    assert uri == uri2



"""Alternate boxing implementation for Weave Trace.

This copies many things from weave/legacy/box.py, but it notably
does not box None and bool which simplify checks for trace users."""

from __future__ import annotations

import datetime
from typing import Any, TypeVar

import numpy as np

from weave.trace.refs import Ref

T = TypeVar("T")


class BoxedInt(int):
    _id: int | None = None
    ref: Ref | None = None


class BoxedFloat(float):
    _id: int | None = None
    ref: Ref | None = None


class BoxedStr(str):
    _id: int | None = None
    ref: Ref | None = None


class BoxedDatetime(datetime.datetime):
    ref: Ref | None = None

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, datetime.datetime)
            and self.timestamp() == other.timestamp()
        )


class BoxedTimedelta(datetime.timedelta):
    ref: Ref | None = None

    def __eq__(self, other: Any) -> bool:
        return (
            isinstance(other, datetime.timedelta)
            and self.total_seconds() == other.total_seconds()
        )


# See https://numpy.org/doc/stable/user/basics.subclassing.html
class BoxedNDArray(np.ndarray):
    ref: Ref | None = None

    def __new__(cls, input_array: Any) -> BoxedNDArray:
        obj = np.asarray(input_array).view(cls)
        return obj

    def __array_finalize__(self, obj: Any) -> None:
        if obj is None:
            return


def box(
    obj: T,
) -> (
    T | BoxedInt | BoxedFloat | BoxedStr | BoxedNDArray | BoxedDatetime | BoxedTimedelta
):
    if type(obj) == int:
        return BoxedInt(obj)
    elif type(obj) == float:
        return BoxedFloat(obj)
    elif type(obj) == str:
        return BoxedStr(obj)
    elif type(obj) == np.ndarray:
        return BoxedNDArray(obj)
    elif type(obj) == datetime.datetime:
        return BoxedDatetime.fromtimestamp(obj.timestamp(), tz=datetime.timezone.utc)
    elif type(obj) == datetime.timedelta:
        return BoxedTimedelta(seconds=obj.total_seconds())
    return obj


def unbox(
    obj: T,
) -> T | int | float | str | np.ndarray | datetime.datetime | datetime.timedelta:
    if type(obj) == BoxedInt:
        return int(obj)
    elif type(obj) == BoxedFloat:
        return float(obj)
    elif type(obj) == BoxedStr:
        return str(obj)
    elif type(obj) == BoxedNDArray:
        return np.array(obj)
    elif type(obj) == BoxedDatetime:
        return datetime.datetime.fromtimestamp(obj.timestamp())
    elif type(obj) == BoxedTimedelta:
        return datetime.timedelta(seconds=obj.total_seconds())
    return obj



"""
This module provides a simple interface to the Sentry SDK.
It is a thin wrapper around the Sentry SDK that provides a few
convenience methods for logging exceptions and marking sessions.
Furthermore, it ensures that the Sentry SDK is properly set up
and that we don't interfere with the user's own Sentry SDK setup.

This file is a trimmed down version of the original WandB Sentry module.
"""

__all__ = ("Sentry",)

import atexit
import functools
import os
import site
import sys
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Type, Union

if sys.version_info >= (3, 8):
    from typing import Literal
else:
    from typing_extensions import Literal

if TYPE_CHECKING:
    from sentry_sdk._types import ExcInfo


import sentry_sdk  # type: ignore
import sentry_sdk.utils  # type: ignore

SENTRY_DEFAULT_DSN = "https://99697cf8ca5158250d3dd6cb23cca9b0@o151352.ingest.us.sentry.io/4507019311251456"

SessionStatus = Literal["ok", "exited", "crashed", "abnormal"]


def _safe_noop(func: Callable) -> Callable:
    """Decorator to ensure that Sentry methods do nothing if disabled and don't raise."""

    @functools.wraps(func)
    def wrapper(self: Type["Sentry"], *args: Any, **kwargs: Any) -> Any:
        if self._disabled:
            return None
        try:
            return func(self, *args, **kwargs)
        except Exception as e:
            # do not call self.exception here to avoid infinite recursion
            if func.__name__ != "exception":
                self.exception(f"Error in {func.__name__}: {e}")
            return None

    return wrapper


class Sentry:
    _disabled: bool

    def __init__(self) -> None:
        self._sent_messages: set = set()

        self.dsn = SENTRY_DEFAULT_DSN

        self.hub: Optional[sentry_sdk.hub.Hub] = None

        self._disabled = False

        # ensure we always end the Sentry session
        atexit.register(self.end_session)

    @property
    def environment(self) -> str:
        """Return the environment we're running in."""
        set_env = os.environ.get("WEAVE_SENTRY_ENV", None)
        if set_env:
            return set_env

        import weave

        is_dev = _is_local_dev_install(weave)
        if is_dev:
            return "development"

        return "production"

    @_safe_noop
    def setup(self) -> None:
        """Setup Sentry SDK.

        We use lower-level APIs (i.e., not sentry_sdk.init) here
        to avoid the possibility of interfering with the user's
        own Sentry SDK setup.
        """
        from .. import version

        client = sentry_sdk.Client(
            dsn=self.dsn,
            default_integrations=False,
            environment=self.environment,
            release=version.VERSION,
        )
        self.hub = sentry_sdk.Hub(client)

    @_safe_noop
    def exception(
        self,
        exc: Union[
            str,
            BaseException,
            "ExcInfo",
            None,
        ],
        handled: bool = False,
        status: Optional["SessionStatus"] = None,
    ) -> None:
        """Log an exception to Sentry."""
        error = Exception(exc) if isinstance(exc, str) else exc
        # based on self.hub.capture_exception(_exc)
        if error is not None:
            exc_info = sentry_sdk.utils.exc_info_from_error(error)
        else:
            exc_info = sys.exc_info()

        event, hint = sentry_sdk.utils.event_from_exception(
            exc_info,
            client_options=self.hub.client.options,  # type: ignore
            mechanism={"type": "generic", "handled": handled},
        )
        try:
            self.hub.capture_event(event, hint=hint)  # type: ignore
        except Exception:
            pass

        # if the status is not explicitly set, we'll set it to "crashed" if the exception
        # was unhandled, or "errored" if it was handled
        status = status or ("crashed" if not handled else "errored")  # type: ignore
        self.mark_session(status=status)

        client, _ = self.hub._stack[-1]  # type: ignore
        if client is not None:
            client.flush()

        return None

    @_safe_noop
    def start_session(self) -> None:
        """Start a new session."""
        assert self.hub is not None
        # get the current client and scope
        _, scope = self.hub._stack[-1]
        session = scope._session

        # if there's no session, start one
        if session is None:
            self.hub.start_session()

    @_safe_noop
    def end_session(self) -> None:
        """End the current session."""
        assert self.hub is not None
        # get the current client and scope
        client, scope = self.hub._stack[-1]
        session = scope._session

        if session is not None and client is not None:
            self.hub.end_session()
            client.flush()

    @_safe_noop
    def mark_session(self, status: Optional["SessionStatus"] = None) -> None:
        """Mark the current session with a status."""
        assert self.hub is not None
        _, scope = self.hub._stack[-1]
        session = scope._session

        if session is not None:
            session.update(status=status)

    @_safe_noop
    def configure_scope(
        self,
        tags: Optional[Dict[str, Any]] = None,
    ) -> None:
        """Configure the Sentry scope for the current thread.

        This function should be called at the beginning of every thread that
        will send events to Sentry. It sets the tags that will be applied to
        all events sent from this thread. It also tries to start a session
        if one doesn't already exist for this thread.
        """
        assert self.hub is not None

        with self.hub.configure_scope() as scope:
            if tags is not None:
                for tag in tags:
                    val = tags.get(tag, None)
                    if val not in (None, ""):
                        scope.set_tag(tag, val)
                    if tag == "user":
                        scope.user = val

        self.start_session()

    # Not in the original WandB Sentry module
    def watch(self) -> Callable:
        def watch_dec(func: Callable) -> Callable:
            """Decorator to watch a function for exceptions and log them to Sentry."""

            @functools.wraps(func)
            def wrapper(*args: Any, **kwargs: Any) -> Any:
                try:
                    return func(*args, **kwargs)
                except Exception as e:
                    self.exception(e)
                    raise

            return wrapper

        return watch_dec


def _is_local_dev_install(module: Any) -> bool:
    # Check if the __file__ attribute exists
    if hasattr(module, "__file__"):
        module_path = module.__file__
        # Check if the path is within any of the site-packages directories
        for directory in site.getsitepackages():
            if directory in module_path:
                return False
        return True
    else:
        return False


global_trace_sentry = Sentry()
global_trace_sentry.setup()
global_trace_sentry.configure_scope()



import weave
from weave.legacy.weave import dispatch, weave_internal
from weave.legacy.weave import weave_types as types

from weave.legacy.tests.util import weavejs_ops


def test_nested_weavejs_call():
    # Store a weave function that uses weavejs raw ops (like plain 'pick') inside of
    # a const node.
    # This ensures that we correctly compile the 'pick' to the appropriate weave
    # python op.
    array = weave.save([{"v": 1}, {"v": 2}])
    node_with_fn = weave.save(
        {
            "a": weave_internal.define_fn(
                {"x": array.type.object_type},
                lambda x: weavejs_ops.weavejs_pick(x, "v") + 1,
            )
        }
    )
    assert weave.use(array.map(lambda row: node_with_fn["a"](row))) == [2, 3]


def test_resolve_static_branches():
    # This relies on auto-execute (since d['a'] is a node), and tests the
    # resolve_static branches code path.
    d = weave.save({"a": 5})
    l = weave.save([1, 2])
    assert weave.use(l.map(lambda row: row + d["a"])) == [6, 7]


def test_empty_list():
    arr = dispatch.RuntimeConstNode(types.List(types.TypedDict({})), [])
    map_fn = weave_internal.define_fn(
        {"row": arr.type.object_type},
        lambda row: row.merge(
            weave.legacy.weave.ops.dict_(output_classid=row["output_class"].id())
        ),
    )

    assert weave.use(arr.map(map_fn)) == []



import importlib
import typing
from functools import wraps

import weave
from weave.trace.op_extensions.accumulator import add_accumulator
from weave.trace.patcher import MultiPatcher, SymbolPatcher

if typing.TYPE_CHECKING:
    from openai.types.chat import ChatCompletionChunk


def maybe_unwrap_api_response(value: typing.Any) -> typing.Any:
    """If the caller requests a raw response, we unwrap the APIResponse object.
    We take a very conservative approach to only unwrap the types we know about.
    """
    maybe_value: typing.Any = None
    try:
        from openai._legacy_response import LegacyAPIResponse

        if isinstance(value, LegacyAPIResponse):
            maybe_value = value.parse()
    except:
        pass

    try:
        from openai._response import APIResponse

        if isinstance(value, APIResponse):
            maybe_value = value.parse()
    except:
        pass

    try:
        from openai.types.chat import ChatCompletion, ChatCompletionChunk

        if isinstance(maybe_value, (ChatCompletion, ChatCompletionChunk)):
            return maybe_value
    except:
        pass

    return value


def openai_on_finish_post_processor(
    value: typing.Optional["ChatCompletionChunk"],
) -> typing.Optional[typing.Dict]:
    from openai.types.chat import ChatCompletion, ChatCompletionChunk
    from openai.types.chat.chat_completion_chunk import (
        ChoiceDeltaFunctionCall,
        ChoiceDeltaToolCall,
    )
    from openai.types.chat.chat_completion_message import FunctionCall
    from openai.types.chat.chat_completion_message_tool_call import (
        ChatCompletionMessageToolCall,
        Function,
    )

    value = maybe_unwrap_api_response(value)

    def _get_function_call(
        function_call: typing.Optional[ChoiceDeltaFunctionCall],
    ) -> typing.Optional[FunctionCall]:
        if function_call is None:
            return function_call
        if isinstance(function_call, ChoiceDeltaFunctionCall):
            return FunctionCall(
                arguments=function_call.arguments,
                name=function_call.name,
            )
        else:
            return None

    def _get_tool_calls(
        tool_calls: typing.Optional[typing.List[ChoiceDeltaToolCall]],
    ) -> typing.Optional[typing.List[ChatCompletionMessageToolCall]]:
        if tool_calls is None:
            return tool_calls

        _tool_calls = []
        if isinstance(tool_calls, list):
            for tool_call in tool_calls:
                assert isinstance(tool_call, ChoiceDeltaToolCall)
                _tool_calls.append(
                    ChatCompletionMessageToolCall(
                        id=tool_call.id,
                        type=tool_call.type,
                        function=Function(
                            name=tool_call.function.name,
                            arguments=tool_call.function.arguments,
                        ),
                    )
                )
        return _tool_calls

    if isinstance(value, ChatCompletionChunk):
        final_value = ChatCompletion(
            id=value.id,
            choices=[
                {
                    "index": choice.index,
                    "message": {
                        "content": choice.delta.content,
                        "role": choice.delta.role,
                        "function_call": _get_function_call(choice.delta.function_call),
                        "tool_calls": _get_tool_calls(choice.delta.tool_calls),
                    },
                    "logprobs": choice.logprobs,
                    "finish_reason": choice.finish_reason,
                }
                for choice in value.choices
            ],
            created=value.created,
            model=value.model,
            object="chat.completion",
            system_fingerprint=value.system_fingerprint,
            usage=value.usage if hasattr(value, "usage") else None,
        )
        return final_value.model_dump(exclude_unset=True, exclude_none=True)
    else:
        return value


def openai_accumulator(
    acc: typing.Optional["ChatCompletionChunk"],
    value: "ChatCompletionChunk",
    skip_last: bool = False,
) -> "ChatCompletionChunk":
    from openai.types.chat import ChatCompletionChunk
    from openai.types.chat.chat_completion_chunk import (
        ChoiceDeltaFunctionCall,
        ChoiceDeltaToolCall,
        ChoiceDeltaToolCallFunction,
    )

    def _process_chunk(
        chunk: ChatCompletionChunk, acc_choices: list[dict] = []
    ) -> typing.List[typing.Dict]:
        """Once the first_chunk is set (acc), take the next chunk and append the message content
        to the message content of acc or first_chunk.
        """
        for chunk_choice in chunk.choices:
            for i in range(chunk_choice.index + 1 - len(acc_choices)):
                acc_choices.append(
                    {
                        "index": len(acc_choices),
                        "delta": {
                            "content": None,
                            "function_call": None,
                            "tool_calls": None,
                        },
                        "logprobs": None,
                        "finish_reason": None,
                    }
                )
            # choice fields
            choice = acc_choices[chunk_choice.index]
            if chunk_choice.finish_reason:
                choice["finish_reason"] = chunk_choice.finish_reason
            if chunk_choice.logprobs:
                choice["logprobs"] = chunk_choice.logprobs

            # message
            if chunk_choice.delta.content:
                if choice["delta"]["content"] is None:
                    choice["delta"]["content"] = ""
                choice["delta"]["content"] += chunk_choice.delta.content  # type: ignore
            if chunk_choice.delta.role:
                choice["delta"]["role"] = chunk_choice.delta.role

            # function calling
            if chunk_choice.delta.function_call:
                if choice["delta"]["function_call"] is None:
                    choice["delta"]["function_call"] = ChoiceDeltaFunctionCall(
                        arguments=chunk_choice.delta.function_call.arguments,
                        name=chunk_choice.delta.function_call.name,
                    )
                else:
                    choice["delta"]["function_call"]["arguments"] += (
                        chunk_choice.delta.function_call.arguments
                    )

            # tool calls
            if chunk_choice.delta.tool_calls:
                if choice["delta"]["tool_calls"] is None:
                    choice["delta"]["tool_calls"] = []
                    tool_call_delta = chunk_choice.delta.tool_calls[
                        0
                    ]  # when streaming, we get one
                    choice["delta"]["tool_calls"].append(  # type: ignore
                        ChoiceDeltaToolCall(
                            id=tool_call_delta.id,
                            index=tool_call_delta.index,
                            function=ChoiceDeltaToolCallFunction(
                                name=tool_call_delta.function.name,  # type: ignore
                                arguments="",
                            ),
                            type=tool_call_delta.type,
                        )
                    )
                else:
                    tool_call_delta = chunk_choice.delta.tool_calls[0]
                    if tool_call_delta.index > len(choice["delta"]["tool_calls"]) - 1:
                        choice["delta"]["tool_calls"].append(
                            ChoiceDeltaToolCall(
                                index=tool_call_delta.index,
                                function=ChoiceDeltaToolCallFunction(
                                    name=None,
                                    arguments="",
                                ),
                            ).model_dump()
                        )
                    tool_call = choice["delta"]["tool_calls"][tool_call_delta.index]
                    if tool_call_delta.id is not None:
                        tool_call["id"] = tool_call_delta.id
                    if tool_call_delta.type is not None:
                        tool_call["type"] = tool_call_delta.type
                    if tool_call_delta.function is not None:
                        if tool_call_delta.function.name is not None:
                            tool_call["function"]["name"] = (
                                tool_call_delta.function.name
                            )
                        if tool_call_delta.function.arguments is not None:
                            tool_call["function"]["arguments"] += (
                                tool_call_delta.function.arguments
                            )

        return acc_choices

    if acc is None:
        if hasattr(value, "choices"):
            output_choices = _process_chunk(value)
            acc = ChatCompletionChunk(
                id=value.id,  # Each chunk has the same ID
                choices=output_choices,
                created=value.created,  # Each chunk has the same timestamp
                model=value.model,
                object=value.object,
                system_fingerprint=value.system_fingerprint,
            )
            return acc
        else:
            raise ValueError("Initial event must contain choices")

    output_choices = _process_chunk(
        value, [choice.model_dump() for choice in acc.choices]
    )

    acc = ChatCompletionChunk(
        id=acc.id,
        choices=output_choices,
        created=acc.created,
        model=acc.model,
        object=acc.object,
        system_fingerprint=acc.system_fingerprint,
    )

    # add usage info
    if len(value.choices) == 0 and value.usage:
        acc.usage = value.usage
        if skip_last:
            raise StopIteration(acc)

    return acc


# Unlike other integrations, streaming is based on input flag
def should_use_accumulator(inputs: typing.Dict) -> bool:
    return (
        isinstance(inputs, dict)
        and bool(inputs.get("stream"))
        # This is very critical. When `"X-Stainless-Raw-Response` is true, the response
        # is an APIResponse object. This is very hard to mock/patch for the streaming use
        # case, so we don't even try.
        and not inputs.get("extra_headers", {}).get("X-Stainless-Raw-Response")
        == "true"
    )


def create_wrapper_sync(
    name: str,
) -> typing.Callable[[typing.Callable], typing.Callable]:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        "We need to do this so we can check if `stream` is used"

        def _add_stream_options(fn: typing.Callable) -> typing.Callable:
            @wraps(fn)
            def _wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                if bool(kwargs.get("stream")) and kwargs.get("stream_options") is None:
                    kwargs["stream_options"] = {"include_usage": True}
                return fn(
                    *args, **kwargs
                )  # This is where the final execution of fn is happening.

            return _wrapper

        def _openai_stream_options_is_set(inputs: typing.Dict) -> bool:
            if inputs.get("stream_options") is not None:
                return True
            return False

        op = weave.op()(_add_stream_options(fn))
        op.name = name  # type: ignore
        return add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: lambda acc, value: openai_accumulator(
                acc, value, skip_last=not _openai_stream_options_is_set(inputs)
            ),
            should_accumulate=should_use_accumulator,
            on_finish_post_processor=openai_on_finish_post_processor,
        )

    return wrapper


# Surprisingly, the async `client.chat.completions.create` does not pass
# `inspect.iscoroutinefunction`, so we can't dispatch on it and must write
# it manually here...
def create_wrapper_async(
    name: str,
) -> typing.Callable[[typing.Callable], typing.Callable]:
    def wrapper(fn: typing.Callable) -> typing.Callable:
        "We need to do this so we can check if `stream` is used"

        def _add_stream_options(fn: typing.Callable) -> typing.Callable:
            @wraps(fn)
            async def _wrapper(*args: typing.Any, **kwargs: typing.Any) -> typing.Any:
                if bool(kwargs.get("stream")) and kwargs.get("stream_options") is None:
                    kwargs["stream_options"] = {"include_usage": True}
                return await fn(*args, **kwargs)

            return _wrapper

        def _openai_stream_options_is_set(inputs: typing.Dict) -> bool:
            if inputs.get("stream_options") is not None:
                return True
            return False

        op = weave.op()(_add_stream_options(fn))
        op.name = name  # type: ignore
        return add_accumulator(
            op,  # type: ignore
            make_accumulator=lambda inputs: lambda acc, value: openai_accumulator(
                acc, value, skip_last=not _openai_stream_options_is_set(inputs)
            ),
            should_accumulate=should_use_accumulator,
            on_finish_post_processor=openai_on_finish_post_processor,
        )

    return wrapper


symbol_patchers = [
    # Patch the Completions.create method
    SymbolPatcher(
        lambda: importlib.import_module("openai.resources.chat.completions"),
        "Completions.create",
        create_wrapper_sync(name="openai.chat.completions.create"),
    ),
    SymbolPatcher(
        lambda: importlib.import_module("openai.resources.chat.completions"),
        "AsyncCompletions.create",
        create_wrapper_async(name="openai.chat.completions.create"),
    ),
]

openai_patcher = MultiPatcher(symbol_patchers)  # type: ignore



import os
from importlib.metadata import version
from typing import Any, Generator

import litellm
import pytest
import semver

import weave

from .litellm import litellm_patcher

# This PR:
# https://github.com/BerriAI/litellm/commit/fe2aa706e8ff4edbcd109897e5da6b83ef6ad693
# Changed the output format for OpenAI to use APIResponse.
# We can handle this in non-streaming mode, but in streaming mode, we
# have no way of correctly capturing the output and not messing up the
# users' code (that i can see). In these cases, model cost is not captured.
USES_RAW_OPENAI_RESPONSE = semver.compare(version("litellm"), "1.42.11") > 0


class Nearly:
    def __init__(self, v: float) -> None:
        self.v = v

    def __eq__(self, other: Any) -> bool:
        return abs(self.v - other) < 2


@pytest.fixture(scope="package")
def patch_litellm(request: Any) -> Generator[None, None, None]:
    # This little hack is to allow us to run the tests in prod mode
    # For some reason pytest's import procedure causes the patching
    # to fail in prod mode. Specifically, the patches get run twice
    # despite the fact that the patcher is a singleton.
    weave_server_flag = request.config.getoption("--weave-server")
    if weave_server_flag == ("prod"):
        yield
        return

    litellm_patcher.attempt_patch()
    yield
    litellm_patcher.undo_patch()


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_litellm_quickstart(
    client: weave.trace.weave_client.WeaveClient, patch_litellm: None
) -> None:
    # This is taken directly from https://docs.litellm.ai/docs/
    chat_response = litellm.completion(
        api_key=os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY"),
        model="gpt-3.5-turbo-0125",
        messages=[{"content": "Hello, how are you?", "role": "user"}],
    )

    all_content = chat_response.choices[0].message.content
    exp = """Hello! I'm just a computer program, so I don't have feelings, but I'm here to help you. How can I assist you today?"""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 2
    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chat_response.id
    assert output["model"] == chat_response.model
    assert output["object"] == chat_response.object
    assert output["created"] == Nearly(chat_response.created)
    summary = call.summary
    assert summary is not None
    model_usage = summary["usage"][output["model"]]
    assert model_usage["requests"] == 1
    assert (
        output["usage"]["completion_tokens"] == model_usage["completion_tokens"] == 31
    )
    assert output["usage"]["prompt_tokens"] == model_usage["prompt_tokens"] == 13
    assert output["usage"]["total_tokens"] == model_usage["total_tokens"] == 44


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_litellm_quickstart_async(
    client: weave.trace.weave_client.WeaveClient, patch_litellm: None
) -> None:
    # This is taken directly from https://docs.litellm.ai/docs/
    chat_response = await litellm.acompletion(
        api_key=os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY"),
        model="gpt-3.5-turbo-0125",
        messages=[{"content": "Hello, how are you?", "role": "user"}],
    )

    all_content = chat_response.choices[0].message.content
    exp = """Hello! I'm just a computer program, so I don't have feelings, but I'm here to help you with whatever you need. How can I assist you today?"""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 2
    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chat_response.id
    assert output["model"] == chat_response.model
    assert output["object"] == chat_response.object
    assert output["created"] == Nearly(chat_response.created)
    summary = call.summary
    assert summary is not None

    model_usage = summary["usage"][output["model"]]
    assert model_usage["requests"] == 1
    assert (
        output["usage"]["completion_tokens"] == model_usage["completion_tokens"] == 35
    )
    assert output["usage"]["prompt_tokens"] == model_usage["prompt_tokens"] == 13
    assert output["usage"]["total_tokens"] == model_usage["total_tokens"] == 48


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
def test_litellm_quickstart_stream(
    client: weave.trace.weave_client.WeaveClient, patch_litellm: None
) -> None:
    # This is taken directly from https://docs.litellm.ai/docs/
    chat_response = litellm.completion(
        api_key=os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY"),
        model="gpt-3.5-turbo-0125",
        messages=[{"content": "Hello, how are you?", "role": "user"}],
        stream=True,
    )

    all_content = ""
    for chunk in chat_response:
        if chunk.choices[0].delta.content:
            all_content += chunk.choices[0].delta.content

    exp = """Hello! I'm just a computer program, so I don't have feelings, but I'm here to help you. How can I assist you today?"""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 2
    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chunk.id
    assert output["model"] == chunk.model
    assert output["created"] == Nearly(chunk.created)
    summary = call.summary
    assert summary is not None

    # We are stuck here:
    # 1. LiteLLM uses raw responses, which we can't wrap in our iterator
    # 2. They don't even capture token usage correctly, so this info is
    # not available for now.
    if not USES_RAW_OPENAI_RESPONSE:
        model_usage = summary["usage"][output["model"]]
        assert model_usage["requests"] == 1
        assert model_usage["completion_tokens"] == 31
        assert model_usage["prompt_tokens"] == 13
        assert model_usage["total_tokens"] == 44


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"], allowed_hosts=["api.wandb.ai", "localhost"]
)
@pytest.mark.asyncio
async def test_litellm_quickstart_stream_async(
    client: weave.trace.weave_client.WeaveClient, patch_litellm: None
) -> None:
    # This is taken directly from https://docs.litellm.ai/docs/
    chat_response = await litellm.acompletion(
        api_key=os.environ.get("OPENAI_API_KEY", "DUMMY_API_KEY"),
        model="gpt-3.5-turbo-0125",
        messages=[{"content": "Hello, how are you?", "role": "user"}],
        stream=True,
    )

    all_content = ""
    async for chunk in chat_response:
        if chunk.choices[0].delta.content:
            all_content += chunk.choices[0].delta.content
    exp = """Hello! I'm just a computer program, so I don't have feelings, but I'm here and ready to assist you with any questions or tasks you may have. How can I help you today?"""

    assert all_content == exp
    calls = list(client.calls())
    assert len(calls) == 2
    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output["choices"][0]["message"]["content"] == exp
    assert output["choices"][0]["finish_reason"] == "stop"
    assert output["id"] == chunk.id
    assert output["model"] == chunk.model
    assert output["created"] == Nearly(chunk.created)
    summary = call.summary
    assert summary is not None

    # We are stuck here:
    # 1. LiteLLM uses raw responses, which we can't wrap in our iterator
    # 2. They don't even capture token usage correctly, so this info is
    # not available for now.
    if not USES_RAW_OPENAI_RESPONSE:
        model_usage = summary["usage"][output["model"]]
        assert model_usage["requests"] == 1
        assert model_usage["completion_tokens"] == 41
        assert model_usage["prompt_tokens"] == 13
        assert model_usage["total_tokens"] == 54


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost"],
)
@pytest.mark.asyncio
def test_model_predict(
    client: weave.trace.weave_client.WeaveClient, patch_litellm: None
) -> None:
    class TranslatorModel(weave.Model):
        model: str
        temperature: float

        @weave.op()
        def predict(self, text: str, target_language: str) -> str:
            response = litellm.completion(
                api_key=os.environ.get("ANTHROPIC_API_KEY", "sk-ant-DUMMY_API_KEY"),
                model=self.model,
                messages=[
                    {
                        "role": "system",
                        "content": f"You are a translator. Translate the given text to {target_language}.",
                    },
                    {"role": "user", "content": text},
                ],
                max_tokens=1024,
                temperature=self.temperature,
            )
            return response.choices[0].message.content

    # Create instances with different models
    claude_translator = TranslatorModel(
        model="claude-3-5-sonnet-20240620", temperature=0.1
    )

    res = claude_translator.predict("There is a bug in my code!", "Spanish")
    assert res is not None

    call = claude_translator.predict.calls()[0]  # type: ignore
    assert dict(call.summary["usage"]["claude-3-5-sonnet-20240620"]) == {
        "requests": 1,
        "prompt_tokens": 28,
        "completion_tokens": 10,
        "total_tokens": 38,
    }



from weave.legacy.weave import weave_types as types
from weave.legacy.weave.decorator_op import op


@op(
    name="isNone",
    input_type={"val": types.optional(types.Any())},
    output_type=types.Boolean(),
)
def is_none(val):
    return val == None



import numpy as np
import pytest

from weave.legacy.weave.ops_primitives import projection_utils

from ... import errors


def test_projection_timeout():
    rng = np.random.RandomState(0)
    embeddings = rng.normal(0, 1, (1000, 50))
    result = projection_utils.perform_2D_projection_with_timeout(
        embeddings, "pca", {}, timeout=1e-6
    )
    assert np.isclose(result, 0).all()


def test_projection_timeout_error_propagation():
    rng = np.random.RandomState(0)
    embeddings = rng.normal(0, 1, (3, 2, 4))
    with pytest.raises(errors.WeaveInternalError):
        projection_utils.perform_2D_projection_with_timeout(
            embeddings, "pca", {}, timeout=1e4
        )



import typing

from weave.legacy.weave import errors, graph, op_args, val_const

from weave.legacy.weave import weave_internal
from weave.legacy.weave import weave_types as types

if typing.TYPE_CHECKING:
    from weave.legacy.weave import op_def


def verify_weave_fn_is_valid(op: "op_def.OpDef", weavified: graph.Node) -> None:
    constituent_ops = typing.cast(
        list[graph.OutputNode],
        graph.filter_nodes_top_level(
            [weavified], lambda n: isinstance(n, graph.OutputNode)
        ),
    )
    output_nodes_disjoint = all([n.from_op.name != op.name for n in constituent_ops])

    if not output_nodes_disjoint:
        raise errors.WeavifyError(
            f"Weavified version of {op.name} contains a call to itself"
        )


def op_to_weave_fn(opdef: "op_def.OpDef") -> graph.Node:
    if opdef.name.startswith("mapped"):
        raise errors.WeavifyError("Cannot convert mapped op to weave_fn here")

    if opdef.name.startswith("Arrow"):
        raise errors.WeavifyError(
            "Refusing to convert op that is already defined on Arrow object to weave_fn"
        )

    if opdef._gets_tag_by_name != None:
        # This happens in test_wb_tables.py::test_join_group_combo. We fail to find the
        # vectorized version because the op is generated in the test and is not a built-in
        raise errors.WeavifyError(
            "Not weavifying tag op. We should already have a vectorized version!"
        )

    if opdef.name.startswith("objectConstructor"):
        raise errors.WeavifyError(
            "Can't convert objectConstructor to weave_fn: %s" % opdef
        )

    # `merge` is a poison pill. It contains a {**dict1, **dict2} operation,
    # which, when dict[1/2] are nodes, never returns - it creates an infinite
    # loop. This will be true for any op def that contains a `**` operation
    # since `{**ops.dict_(a=1)}` will just hang forever.
    if opdef.name == "merge":
        raise errors.WeavifyError("Can't convert merge to weave_fn")

    # if is_base_op(opdef):
    #     raise errors.WeaveTypeError("Refusing to convert base op to weave_fn.")

    input_type = opdef.input_type
    if isinstance(input_type, op_args.OpVarArgs):
        raise errors.WeavifyError("Refusing to convert variadic op to weave_fn.")
    input_type = typing.cast(op_args.OpNamedArgs, input_type)

    if any(types.is_custom_type(t) for t in input_type.arg_types.values()):
        raise errors.WeavifyError("Can't weavify op with custom typed args")

    original_input_types = typing.cast(
        types.TypedDict, input_type.weave_type()
    ).property_types

    def weave_fn_body(*args: graph.VarNode) -> graph.Node:
        kwargs = {key: args[i] for i, key in enumerate(original_input_types)}
        result = opdef.resolve_fn(**kwargs)  # type: ignore
        return weavify_object(result)

    fn = weave_internal.define_fn(original_input_types, weave_fn_body).val
    verify_weave_fn_is_valid(opdef, fn)

    return fn


def weavify_object(obj: typing.Any) -> graph.Node:
    from weave.legacy.weave.ops_primitives import dict_, make_list

    if isinstance(obj, graph.Node):
        return obj
    elif isinstance(obj, val_const.Const):
        return weave_internal.make_const_node(
            types.Const(types.TypeRegistry.type_of(obj.val), obj.val), obj.val
        )
    elif isinstance(obj, list):
        return make_list(**{str(i): weavify_object(o) for i, o in enumerate(obj)})
    elif isinstance(obj, dict):
        return dict_(**{i: weavify_object(o) for i, o in obj.items()})

    # ObjectType constructor is disabled.
    # try:
    #     return obj.__class__.constructor(
    #         **{
    #             field.name: weavify_object(getattr(obj, field.name))
    #             for field in dataclasses.fields(obj)
    #         }
    #     )
    # except AttributeError:
    #     pass
    return weave_internal.make_const_node(types.TypeRegistry.type_of(obj), obj)



import asyncio
import os
from typing import Union

import pytest

import weave
from weave.integrations.integration_utilities import (
    flatten_calls,
    flattened_calls_to_names,
)
from weave.trace_server.trace_server_interface import CallsFilter


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
)
def test_groq_quickstart(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from groq import Groq

    groq_client = Groq(
        api_key=os.environ.get("GROQ_API_KEY", "DUMMY_API_KEY"),
    )
    chat_completion = groq_client.chat.completions.create(
        messages=[
            {
                "role": "user",
                "content": "What is the capital of India?",
            }
        ],
        model="llama3-8b-8192",
        seed=42,
    )

    assert (
        chat_completion.choices[0].message.content
        == "The capital of India is New Delhi."
    )
    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert len(flattened_calls) == 1

    assert flattened_calls_to_names(flattened_calls) == [
        ("groq.chat.completions.create", 0),
    ]

    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.id == chat_completion.id
    assert output.model == chat_completion.model
    assert output.usage.completion_tokens == 9
    assert output.usage.prompt_tokens == 17
    assert output.usage.total_tokens == 26
    assert output.choices[0].finish_reason == "stop"
    assert output.choices[0].message.content == "The capital of India is New Delhi."


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
)
def test_groq_async_chat_completion(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from groq import AsyncGroq

    groq_client = AsyncGroq(api_key=os.environ.get("GROQ_API_KEY", "DUMMY_API_KEY"))

    async def complete_chat() -> None:
        chat_completion = await groq_client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": "You are a psychiatrist helping young minds",
                },
                {
                    "role": "user",
                    "content": "I panicked during the test, even though I knew everything on the test paper.",
                },
            ],
            model="llama3-8b-8192",
            temperature=0.3,
            max_tokens=360,
            top_p=1,
            stop=None,
            stream=False,
            seed=42,
        )

    asyncio.run(complete_chat())

    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert len(flattened_calls) == 1

    assert flattened_calls_to_names(flattened_calls) == [
        ("groq.async.chat.completions.create", 0),
    ]

    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.model == "llama3-8b-8192"
    assert output.usage.completion_tokens == 152
    assert output.usage.prompt_tokens == 38
    assert output.usage.total_tokens == 190
    assert output.choices[0].finish_reason == "stop"
    assert (
        output.choices[0].message.content
        == """It sounds like you're feeling really frustrated and disappointed with yourself right now. It's completely normal to feel that way, especially when you're used to performing well and then suddenly feel like you've let yourself down.

Can you tell me more about what happened during the test? What was going through your mind when you started to feel panicked? Was it the pressure of the test itself, or was there something else that triggered your anxiety?

Also, have you experienced panic or anxiety during tests before? If so, what strategies have you used to cope with those feelings in the past?

Remember, as your psychiatrist, my goal is to help you understand what's going on and find ways to manage your anxiety so you can perform to the best of your ability."""
    )


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
)
def test_groq_streaming_chat_completion(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from groq import Groq

    groq_client = Groq(api_key=os.environ.get("GROQ_API_KEY", "DUMMY_API_KEY"))

    stream = groq_client.chat.completions.create(
        messages=[
            {"role": "system", "content": "you are a helpful assistant."},
            {
                "role": "user",
                "content": "Explain the importance of fast language models",
            },
        ],
        model="llama3-8b-8192",
        temperature=0.5,
        max_tokens=1024,
        top_p=1,
        stop=None,
        stream=True,
        seed=42,
    )

    all_content = ""
    for chunk in stream:
        if chunk.choices[0].delta.content is not None:
            all_content += chunk.choices[0].delta.content

    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert len(flattened_calls) == 1

    assert flattened_calls_to_names(flattened_calls) == [
        ("groq.chat.completions.create", 0),
    ]

    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.model == "llama3-8b-8192"
    assert output.object == "chat.completion"
    assert output.usage.completion_tokens == 533
    assert output.usage.prompt_tokens == 29
    assert output.usage.total_tokens == 562
    assert output.usage.completion_time > 0
    assert output.usage.prompt_time > 0
    assert output.usage.queue_time > 0
    assert output.usage.total_time > 0

    assert len(output.choices) == 1
    assert output.choices[0].finish_reason == "stop"
    assert output.choices[0].index == 0
    assert output.choices[0].message.role == "assistant"
    assert (
        output.choices[0].message.content
        == """Fast language models have gained significant attention in recent years due to their ability to process and generate human-like language at incredibly high speeds. Here are some reasons why fast language models are important:

1. **Real-time Applications**: Fast language models can be used in real-time applications such as chatbots, virtual assistants, and language translation systems. They can quickly respond to user queries, making them more interactive and engaging.
2. **Efficient Processing**: Fast language models can process large amounts of text data quickly, making them ideal for tasks such as sentiment analysis, text classification, and topic modeling. This efficiency is particularly important in applications where speed is critical, such as in customer service or emergency response systems.
3. **Improved Responsiveness**: Fast language models can respond quickly to user input, reducing the latency and improving the overall user experience. This is particularly important in applications where users expect instant responses, such as in gaming or social media platforms.
4. **Scalability**: Fast language models can handle large volumes of data and scale up or down as needed, making them suitable for applications with varying traffic patterns.
5. **Advancements in AI Research**: Fast language models have enabled researchers to explore new areas of natural language processing (NLP), such as language generation, question answering, and dialogue systems. This has led to significant advancements in AI research and the development of more sophisticated language models.
6. **Improved Language Understanding**: Fast language models can be fine-tuned for specific tasks, such as named entity recognition, part-of-speech tagging, and dependency parsing. This has led to improved language understanding and better performance in various NLP tasks.
7. **Enhanced User Experience**: Fast language models can be used to create more personalized and engaging user experiences, such as recommending products or services based on user preferences and behavior.
8. **Cost-Effective**: Fast language models can be more cost-effective than traditional language models, as they require less computational resources and can be deployed on cloud-based infrastructure.
9. **Faster Development Cycles**: Fast language models can accelerate the development cycle of language-based applications, enabling developers to iterate and refine their models more quickly.
10. **Broader Adoption**: Fast language models have made NLP more accessible to a broader range of developers and organizations, enabling them to build language-based applications without requiring extensive expertise in NLP.

In summary, fast language models have revolutionized the field of NLP, enabling the development of more efficient, scalable, and responsive language-based applications. Their importance lies in their ability to process and generate human-like language at incredible speeds, making them a crucial component of many modern AI systems."""
    )


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
)
def test_groq_async_streaming_chat_completion(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    from groq import AsyncGroq

    groq_client = AsyncGroq(api_key=os.environ.get("GROQ_API_KEY", "DUMMY_API_KEY"))

    async def generate_reponse() -> str:
        chat_streaming = await groq_client.chat.completions.create(
            messages=[
                {
                    "role": "system",
                    "content": "You are a psychiatrist helping young minds",
                },
                {
                    "role": "user",
                    "content": "I panicked during the test, even though I knew everything on the test paper.",
                },
            ],
            model="llama3-8b-8192",
            temperature=0.3,
            max_tokens=360,
            top_p=1,
            stop=None,
            stream=True,
            seed=42,
        )

        all_content = ""
        async for chunk in chat_streaming:
            if chunk.choices[0].delta.content is not None:
                all_content += chunk.choices[0].delta.content

        return all_content

    asyncio.run(generate_reponse())

    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert len(flattened_calls) == 1

    assert flattened_calls_to_names(flattened_calls) == [
        ("groq.async.chat.completions.create", 0),
    ]

    call = calls[0]
    assert call.exception is None and call.ended_at is not None
    output = call.output
    assert output.model == "llama3-8b-8192"
    assert output.usage.completion_tokens == 152
    assert output.usage.prompt_tokens == 38
    assert output.usage.total_tokens == 190
    assert output.usage.completion_time > 0
    assert output.usage.prompt_time > 0
    assert output.usage.queue_time > 0
    assert output.usage.total_time > 0
    assert output.choices[0].finish_reason == "stop"
    assert (
        output.choices[0].message.content
        == """It sounds like you're feeling really frustrated and disappointed with yourself right now. It's completely normal to feel that way, especially when you're used to performing well and then suddenly feel like you've let yourself down.

Can you tell me more about what happened during the test? What was going through your mind when you started to feel panicked? Was it the pressure of the test itself, or was there something else that triggered your anxiety?

Also, have you experienced panic or anxiety during tests before? If so, what strategies have you used to cope with those feelings in the past?

Remember, as your psychiatrist, my goal is to help you understand what's going on and find ways to manage your anxiety so you can perform to the best of your ability."""
    )


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization", "x-api-key"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
)
def test_groq_tool_call(
    client: weave.trace.weave_client.WeaveClient,
) -> None:
    import json

    from groq import Groq

    groq_client = Groq(api_key=os.getenv("GROQ_API_KEY", "DUMMY_API_KEY"))

    @weave.op()
    def get_game_score(team_name: str) -> str:
        """Get the current score for a given NBA game"""
        if "warriors" in team_name.lower():
            return json.dumps(
                {
                    "game_id": "401585601",
                    "status": "Final",
                    "home_team": "Los Angeles Lakers",
                    "home_team_score": 121,
                    "away_team": "Golden State Warriors",
                    "away_team_score": 128,
                }
            )
        elif "lakers" in team_name.lower():
            return json.dumps(
                {
                    "game_id": "401585601",
                    "status": "Final",
                    "home_team": "Los Angeles Lakers",
                    "home_team_score": 121,
                    "away_team": "Golden State Warriors",
                    "away_team_score": 128,
                }
            )
        elif "nuggets" in team_name.lower():
            return json.dumps(
                {
                    "game_id": "401585577",
                    "status": "Final",
                    "home_team": "Miami Heat",
                    "home_team_score": 88,
                    "away_team": "Denver Nuggets",
                    "away_team_score": 100,
                }
            )
        elif "heat" in team_name.lower():
            return json.dumps(
                {
                    "game_id": "401585577",
                    "status": "Final",
                    "home_team": "Miami Heat",
                    "home_team_score": 88,
                    "away_team": "Denver Nuggets",
                    "away_team_score": 100,
                }
            )
        else:
            return json.dumps({"team_name": team_name, "score": "unknown"})

    @weave.op()
    def run_conversation(user_prompt: str) -> Union[str, None]:
        # Step 1: send the conversation and available functions to the model
        messages = [
            {
                "role": "system",
                "content": "You are a function calling LLM that uses the data extracted from the get_game_score function to answer questions around NBA game scores. Include the team and their opponent in your response.",
            },
            {
                "role": "user",
                "content": user_prompt,
            },
        ]
        tools = [
            {
                "type": "function",
                "function": {
                    "name": "get_game_score",
                    "description": "Get the score for a given NBA game",
                    "parameters": {
                        "type": "object",
                        "properties": {
                            "team_name": {
                                "type": "string",
                                "description": "The name of the NBA team (e.g. 'Golden State Warriors')",
                            }
                        },
                        "required": ["team_name"],
                    },
                },
            }
        ]
        response = groq_client.chat.completions.create(
            model="llama3-70b-8192",
            messages=messages,
            tools=tools,
            tool_choice="auto",
            max_tokens=4096,
            seed=42,
        )

        response_message = response.choices[0].message
        tool_calls = response_message.tool_calls
        # Step 2: check if the model wanted to call a function
        if tool_calls:
            # Step 3: call the function
            # Note: the JSON response may not always be valid; be sure to handle errors
            available_functions = {
                "get_game_score": get_game_score,
            }  # only one function in this example, but you can have multiple
            messages.append(
                response_message
            )  # extend conversation with assistant's reply
            # Step 4: send the info for each function call and function response to the model
            for tool_call in tool_calls:
                function_name = tool_call.function.name
                function_to_call = available_functions[function_name]
                function_args = json.loads(tool_call.function.arguments)
                function_response = function_to_call(
                    team_name=function_args.get("team_name")
                )
                messages.append(
                    {
                        "tool_call_id": tool_call.id,
                        "role": "tool",
                        "name": function_name,
                        "content": function_response,
                    }
                )  # extend conversation with function response
            second_response = groq_client.chat.completions.create(
                model="llama3-70b-8192", messages=messages, seed=42
            )  # get a new response from the model where it can see the function response
            return second_response.choices[0].message.content
        return None

    response = run_conversation("What was the score of the Warriors game?")

    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert len(flattened_calls) == 4

    assert flattened_calls_to_names(flattened_calls) == [
        ("run_conversation", 0),
        ("groq.chat.completions.create", 1),
        ("get_game_score", 1),
        ("groq.chat.completions.create", 1),
    ]

    call_0, _ = flattened_calls[0]
    assert call_0.exception is None and call_0.ended_at is not None
    output_0 = call_0.output
    assert output_0 == response

    call_1, _ = flattened_calls[1]
    assert call_1.exception is None and call_1.ended_at is not None
    output_1 = call_1.output
    assert output_1.usage.completion_tokens == 47
    assert output_1.usage.prompt_tokens == 973
    assert output_1.usage.total_tokens == 1020
    assert output_1.usage.completion_time > 0
    assert output_1.usage.prompt_time > 0
    assert output_1.usage.total_time > 0
    assert output_1.model == "llama3-70b-8192"
    assert len(output_1.choices) == 1
    assert output_1.choices[0].finish_reason == "tool_calls"
    assert output_1.choices[0].index == 0
    assert output_1.choices[0].message.role == "assistant"
    assert len(output_1.choices[0].message.tool_calls) == 1
    assert (
        output_1.choices[0].message.tool_calls[0].function.arguments
        == '{"team_name":"Golden State Warriors"}'
    )
    assert output_1.choices[0].message.tool_calls[0].type == "function"

    call_2, _ = flattened_calls[2]
    assert call_2.exception is None and call_2.ended_at is not None
    output_2 = call_2.output
    game_score_data = json.loads(output_2)
    assert game_score_data["game_id"] == "401585601"
    assert game_score_data["status"] == "Final"
    assert game_score_data["home_team"] == "Los Angeles Lakers"
    assert game_score_data["home_team_score"] == 121
    assert game_score_data["away_team"] == "Golden State Warriors"
    assert game_score_data["away_team_score"] == 128

    call_3, _ = flattened_calls[3]
    assert call_3.exception is None and call_3.ended_at is not None
    output_3 = call_3.output
    assert output_3.usage.completion_tokens == 20
    assert output_3.usage.prompt_tokens == 177
    assert output_3.usage.total_tokens == 197
    assert output_3.usage.completion_time > 0
    assert output_3.usage.prompt_time > 0
    assert output_3.usage.total_time > 0
    assert output_3.model == "llama3-70b-8192"
    assert len(output_3.choices) == 1
    assert output_3.choices[0].finish_reason == "stop"
    assert output_3.choices[0].index == 0
    assert output_3.choices[0].message.role == "assistant"
    assert (
        output_3.choices[0].message.content
        == "The Golden State Warriors played against the Los Angeles Lakers and won the game 128-121."
    )



# Automatic op testing.
#
# This automatically tests all ops defined in op_specs.py. We generate a set
# of variant test cases for every op. It is easy to add checks like
# "binary commutative ops should be produce None when vectorized with a null
# scalar as the other argument".

import dataclasses
import itertools
import typing

import pytest

import weave
from weave.legacy.weave import (
    graph,
    graph_debug,
    language_nullability,
    op_def,
    ops_arrow,
    ops_primitives,
    storage,
    weave_internal,
)
from weave.legacy.weave.language_features.tagging import make_tag_getter_op
from weave.legacy.weave.language_features.tagging.tagged_value_type import (
    TaggedValueType,
)

from ...legacy.weave import registry_mem
from weave.legacy.tests.util.concrete_tagged_value import (
    TaggedValue,
    concrete_from_tagstore,
    concrete_to_tagstore,
)
from weave.legacy.tests.util.op_specs import OP_TEST_SPECS, OpSpec, OpSpecTestCase


def assert_equal_with_tags(node: graph.Node, v: typing.Any, expected: typing.Any):
    v = concrete_from_tagstore(v)
    if isinstance(expected, TaggedValue):
        # Nones are not tagged
        if v != None:
            assert isinstance(v, TaggedValue)
            assert expected.tag == v.tag
            v = v.value
        expected = expected.value

    if isinstance(v, ops_arrow.ArrowWeaveList):
        assert isinstance(expected, ops_arrow.ArrowWeaveList)
        actual_py = v.to_pylist_raw()
        expected_py = expected.to_pylist_raw()
        assert actual_py == expected_py
        # We can't check the arrow type if we expect all Nones, because we pass
        # a python list in as the expectation right now. And in arrow you can have
        # Nones of any type.
        if not all(x is None for x in expected):
            assert v._arrow_data.type == expected._arrow_data.type
        # TODO: this check is too loose, it should be type equality. But
        # we currently aren't specific enough in some ops, or have mistmatches
        # between python and arrow
        assert expected.object_type.assign_type(v.object_type)
    else:
        assert v == expected


def assert_fn_vectorized(fn_node: graph.Node) -> graph.Node:
    vec_fn_node = ops_arrow.vectorize(fn_node)

    def assert_node_vectorized(n: graph.Node):
        if isinstance(n, graph.OutputNode):
            assert isinstance(
                n.type, ops_arrow.ArrowWeaveListType
            ), f"fn_node {graph_debug.node_expr_str_full(fn_node)} is not fully vectorized. Vectorized as {graph_debug.node_expr_str_full(vec_fn_node)}. Node {graph_debug.node_expr_str_full(n)} is of type {n.type}. All nodes within a vectorized function should be ArrowWeaveListType."

    graph.map_nodes_top_level([vec_fn_node], assert_node_vectorized)
    return vec_fn_node


def check_case(called: graph.Node, result_type: weave.types.Type, result: typing.Any):
    assert result_type.assign_type(
        called.type
    ), f"Expected op output type: {result_type}, but got {called.type}"

    # This is a way to save the final output, preserving tags.
    result_ref = weave.use(weave.legacy.weave.ops.save_to_ref(called, None))
    if result_ref == None:
        actual_result = None
    else:
        actual_result = storage.get(str(result_ref))
    assert_equal_with_tags(called, actual_result, result)

    # # We can't check the type of the output we received, because execution
    # # doesn't produce tags directly. But since we know result was equal
    # # to expectation from above, we can check that the type of the
    # # expectation instead.
    actual_result_type = weave.type_of(concrete_to_tagstore(result))
    # # TODO: this check is too loose, it should be type equality. But
    # # we currently aren't specific enough in some ops, or have mistmatches
    # # between python and arrow
    assert result_type.assign_type(
        actual_result_type
    ), f"{graph_debug.node_expr_str_full(called)} produced type {actual_result_type}, but expected type {result_type}"


def call(op_def: op_def.OpDef, args: typing.Any):
    inputs = [concrete_to_tagstore(it) for it in args]
    # We have to save inputs to make the tags available inside the executor.
    # Users will never have to do this since they don't interact with tests
    # directly, but it slows the test down since saving writes to disk.
    saved_inputs = [weave.save(it) if not callable(it) else it for it in inputs]

    return op_def(*saved_inputs)


class NoValue:
    pass


@dataclasses.dataclass
class OpTestCase:
    name: list[str]

    def check(self):
        raise NotImplementedError()


@dataclasses.dataclass
class OpTestCaseStandard(OpTestCase):
    op_def: op_def.OpDef
    input: tuple[typing.Any, ...]
    expected_type: typing.Any
    expected: typing.Any

    def variant(
        self,
        name: str,
        input: typing.Union[NoValue, tuple[typing.Any, ...]] = NoValue(),
        expected_type: typing.Union[NoValue, typing.Any] = NoValue(),
        expected: typing.Union[NoValue, typing.Any] = NoValue(),
    ):
        return OpTestCaseStandard(
            name=self.name + [name],
            op_def=self.op_def,
            input=self.input if isinstance(input, NoValue) else input,
            expected_type=self.expected_type
            if isinstance(expected_type, NoValue)
            else expected_type,
            expected=self.expected if isinstance(expected, NoValue) else expected,
        )

    def vecdim0_variant(
        self,
        name: str,
        input: typing.Union[NoValue, tuple[typing.Any, ...]] = NoValue(),
        expected_type: typing.Union[NoValue, typing.Any] = NoValue(),
        expected: typing.Union[NoValue, typing.Any] = NoValue(),
    ):
        return OpTestCaseVecDim0(
            name=self.name + [name],
            op_def=self.op_def,
            input=self.input if isinstance(input, NoValue) else input,
            expected_type=self.expected_type
            if isinstance(expected_type, NoValue)
            else expected_type,
            expected=self.expected if isinstance(expected, NoValue) else expected,
        )

    def vecdim1_variant(
        self,
        name: str,
        input: typing.Union[NoValue, tuple[typing.Any, ...]] = NoValue(),
        expected_type: typing.Union[NoValue, typing.Any] = NoValue(),
        expected: typing.Union[NoValue, typing.Any] = NoValue(),
    ):
        return OpTestCaseVecDim1(
            name=self.name + [name],
            op_def=self.op_def,
            input=self.input if isinstance(input, NoValue) else input,
            expected_type=self.expected_type
            if isinstance(expected_type, NoValue)
            else expected_type,
            expected=self.expected if isinstance(expected, NoValue) else expected,
        )

    def check(self):
        check_case(
            call(self.op_def, self.input),
            result_type=self.expected_type,
            result=self.expected,
        )


def convert_expected_to_arrow(
    expected_type: weave.types.Type, expected: typing.Any
) -> tuple[weave.types.Type, typing.Any]:
    if isinstance(expected_type, weave.types.List):
        assert isinstance(expected, list)
        arrow_type = ops_arrow.ArrowWeaveListType(expected_type.object_type)
        arrow_list = ops_arrow.to_arrow(concrete_to_tagstore(expected))
        # We override the object type in the expected value, to be the declared expected
        # type. The arrow list value may have a type that is too narrow for checking
        # (right now ArrowWeaveLists constructed inside op resolvers tend to use the
        # input type logic, which creates wider types than doing a type_of operation would)
        arrow_list.object_type = arrow_type.object_type
        return arrow_type, arrow_list
    if isinstance(expected_type, TaggedValueType):
        assert isinstance(expected, TaggedValue)
        expected_value_type, expected_value = convert_expected_to_arrow(
            expected_type.value, expected.value
        )
        return TaggedValueType(expected_type.tag, expected_value_type), TaggedValue(
            expected.tag, expected_value
        )
    return expected_type, expected


@dataclasses.dataclass
class OpTestCaseVecDim0(OpTestCaseStandard):
    """Tests that a list op has a vectorized equivalent."""

    op_def: op_def.OpDef
    input: tuple[typing.Any, ...]
    expected_type: typing.Any
    expected: typing.Any

    def check(self):
        inputs = [concrete_to_tagstore(it) for it in self.input]
        inputs[0] = ops_arrow.to_arrow(inputs[0])
        inputs = [
            weave.save(input) if not callable(input) else input for input in inputs
        ]
        called = getattr(inputs[0], self.op_def.simple_name)(*inputs[1:])
        called_op_name = called.from_op.name
        called_op_def = registry_mem.memory_registry.get_op(called_op_name)
        assert isinstance(
            list(called_op_def.input_type.arg_types.values())[0],
            ops_arrow.ArrowWeaveListType,
        )
        expected_type, expected = convert_expected_to_arrow(
            self.expected_type, self.expected
        )
        check_case(
            called,
            result_type=expected_type,
            result=expected,
        )


@dataclasses.dataclass
class OpTestCaseVecDim1(OpTestCaseStandard):
    """Tests that a list op has a vectorized equivalent for second dim (joinall/concat)."""

    op_def: op_def.OpDef
    input: tuple[typing.Any, ...]
    expected_type: typing.Any
    expected: typing.Any

    def check(self):
        inputs = [concrete_to_tagstore(it) for it in self.input]
        inputs[0] = [ops_arrow.to_arrow(i) for i in inputs[0]]
        if isinstance(self.input[0], TaggedValue):
            inputs[0] = concrete_to_tagstore(TaggedValue(self.input[0].tag, inputs[0]))
        inputs = [
            weave.save(input) if not callable(input) else input for input in inputs
        ]
        called = getattr(inputs[0], self.op_def.simple_name)(*inputs[1:])
        assert ops_arrow.ArrowWeaveListType().assign_type(called.type)
        expected_type, expected = convert_expected_to_arrow(
            self.expected_type, self.expected
        )
        check_case(
            called,
            result_type=expected_type,
            result=expected,
        )


@dataclasses.dataclass
class OpTestCaseVector(OpTestCase):
    """Test there is a vectorized version that handles all inputs as vectors."""

    op_def: op_def.OpDef
    input_vecs: tuple[list[typing.Any], ...]
    expected_object_type: typing.Any
    expected: typing.Any

    def variant(
        self,
        name: str,
        input_vecs: typing.Union[NoValue, tuple[list[typing.Any], ...]] = NoValue(),
        expected_object_type: typing.Union[NoValue, typing.Any] = NoValue(),
        expected: typing.Union[NoValue, typing.Any] = NoValue(),
    ):
        return OpTestCaseVector(
            name=self.name + [name],
            op_def=self.op_def,
            input_vecs=self.input_vecs
            if isinstance(input_vecs, NoValue)
            else input_vecs,
            expected_object_type=self.expected_object_type
            if isinstance(expected_object_type, NoValue)
            else expected_object_type,
            expected=self.expected if isinstance(expected, NoValue) else expected,
        )

    def check(self):
        input_vars = {
            k: weave.type_of(v).object_type
            for k, v in zip(self.op_def.input_type.arg_types, self.input_vecs)
        }
        op_fn = weave_internal.define_fn(input_vars, lambda *args: self.op_def(*args))
        vec_op_fn = assert_fn_vectorized(op_fn.val)
        called = weave_internal.call_fn(
            vec_op_fn,
            {
                k: ops_arrow.to_weave_arrow(v)
                for k, v in zip(input_vars, self.input_vecs)
            },
        )
        check_case(
            called,
            result_type=ops_arrow.ArrowWeaveListType(self.expected_object_type),
            result=ops_arrow.to_arrow(self.expected),
        )


@dataclasses.dataclass
class OpTestCaseBinaryVectorScalar(OpTestCase):
    """Test binary op vectorizes with (vec, scalar) or (scalar, vec) call"""

    op_def: op_def.OpDef
    vec: list[typing.Any]
    scalar: typing.Any
    vec_first: bool
    expected_object_type: typing.Any
    expected: typing.Any

    def variant(
        self,
        name: str,
        vec: typing.Union[NoValue, list[typing.Any]] = NoValue(),
        scalar: typing.Union[NoValue, typing.Any] = NoValue(),
        vec_first: typing.Union[NoValue, bool] = NoValue(),
        expected_object_type: typing.Union[NoValue, typing.Any] = NoValue(),
        expected: typing.Union[NoValue, typing.Any] = NoValue(),
    ):
        return OpTestCaseBinaryVectorScalar(
            name=self.name + [name],
            op_def=self.op_def,
            vec=self.vec if isinstance(vec, NoValue) else vec,
            scalar=self.scalar if isinstance(scalar, NoValue) else scalar,
            vec_first=self.vec_first if isinstance(vec_first, NoValue) else vec_first,
            expected_object_type=self.expected_object_type
            if isinstance(expected_object_type, NoValue)
            else expected_object_type,
            expected=self.expected if isinstance(expected, NoValue) else expected,
        )

    def check(self):
        # Check the python listmap version
        map_fn = (
            lambda x: self.op_def(x, self.scalar)
            if self.vec_first
            else self.op_def(self.scalar, x)
        )
        check_case(
            ops_primitives.List._listmap(self.vec, map_fn),
            result_type=weave.types.List(self.expected_object_type),
            result=self.expected,
        )
        # Check arrow version
        arrow_vec = ops_arrow.to_arrow(self.vec)
        called = call(ops_arrow.ops.map, (arrow_vec, map_fn))
        # ensure it vectorized
        assert_fn_vectorized(called.from_op.inputs["map_fn"].val)
        check_case(
            called,
            result_type=ops_arrow.ArrowWeaveListType(self.expected_object_type),
            result=ops_arrow.to_arrow(self.expected),
        )


@dataclasses.dataclass
class OpTestCaseBinaryVectorVector(OpTestCase):
    """Test binary op vectorizes with (a, a) call"""

    op_def: op_def.OpDef
    arity: int
    vec: list[typing.Any]

    def variant(
        self,
        name: str,
        vec: typing.Union[NoValue, list[typing.Any]] = NoValue(),
    ):
        return OpTestCaseBinaryVectorVector(
            name=self.name + [name],
            op_def=self.op_def,
            arity=self.arity,
            vec=self.vec if isinstance(vec, NoValue) else vec,
        )

    def check(self):
        def map_fn(item):
            return self.op_def(*tuple(item for i in range(self.arity)))

        listmap_called = call(ops_primitives.List._listmap, (self.vec, map_fn))
        listmap_result = weave.use(listmap_called)
        vec_called = call(ops_arrow.ops.map, (ops_arrow.to_arrow(self.vec), map_fn))
        assert_fn_vectorized(vec_called.from_op.inputs["map_fn"].val)
        vec_result = weave.use(vec_called)
        assert listmap_result == vec_result.to_pylist_tagged()


def make_standard_variants(
    op_test_def: OpSpec, test_case: OpSpecTestCase, test_case_index: int
) -> list[OpTestCaseStandard]:
    op_name = op_test_def.op.name
    base = OpTestCaseStandard(
        [op_name, f"case{test_case_index}"],
        op_test_def.op,
        test_case.input,
        test_case.expected_type,
        test_case.expected,
    )
    base_variants: list[OpTestCaseStandard] = []
    # Check the case itself
    base_variants.append(base)

    # Check commutativity
    if op_test_def.kind.arity == 2 and op_test_def.kind.commutative:
        base_variants.append(base.variant("commutative", input=test_case.input[::-1]))

    # the op should handle None in any position, returning None.
    is_null_consuming = not language_nullability.should_force_none_result(
        {"first_arg": None}, op_test_def.op
    )

    if not is_null_consuming:
        inputs = list(test_case.input)
        inputs[0] = None

        base_variants.append(
            base.variant(
                "arg0-none",
                input=tuple(inputs),
                expected_type=weave.types.optional(test_case.expected_type),
                expected=None,
            )
        )
        if op_test_def.kind.arity == 2:
            base_variants.append(
                base.variant(
                    "arg1-none",
                    input=(test_case.input[0], None),
                    expected_type=weave.types.optional(test_case.expected_type),
                    expected=None,
                )
            )

    # Tags in first position are flowed to output.
    inputs = list(test_case.input)
    inputs[0] = TaggedValue({"a": 5}, inputs[0])
    base_variants.append(
        base.variant(
            "arg0-tag",
            input=tuple(inputs),
            expected_type=TaggedValueType(
                weave.types.TypedDict({"a": weave.types.Int()}),
                test_case.expected_type,
            ),
            expected=TaggedValue({"a": 5}, test_case.expected),
        )
    )

    if op_test_def.kind.arity == 2:
        # Tags in second position are dropped
        base_variants.append(
            base.variant(
                "rhs-tag",
                input=(
                    test_case.input[0],
                    TaggedValue({"a": 5}, test_case.input[1]),
                ),
            )
        )

    return base_variants


def make_standard_vector_variants(
    op_test_def: OpSpec,
) -> list[OpTestCaseVector]:
    op_name = op_test_def.op.name
    input_vecs = [
        [tc.input[i] for tc in op_test_def.test_cases]
        for i in range(op_test_def.kind.arity)
    ]
    variants = []

    base_variant = OpTestCaseVector(
        [op_name, "vector"],
        op_test_def.op,
        tuple(input_vecs),
        weave.types.union(*[tc.expected_type for tc in op_test_def.test_cases]),
        [tc.expected for tc in op_test_def.test_cases],
    )
    variants.append(base_variant)

    input_vecs_with_arg0_nones = []
    input_vecs_with_arg0_nones.append(input_vecs[0] + [None] * len(input_vecs[0]))
    if op_test_def.kind.arity == 2:
        input_vecs_with_arg0_nones.append(input_vecs[1] * 2)

    is_null_consuming = not language_nullability.should_force_none_result(
        {"first_arg": None}, op_test_def.op
    )

    if not is_null_consuming:
        variants.append(
            base_variant.variant(
                "arg0-none",
                input_vecs=tuple(input_vecs_with_arg0_nones),
                expected_object_type=weave.types.optional(
                    base_variant.expected_object_type
                ),
                expected=base_variant.expected + [None] * len(input_vecs[0]),
            )
        )

        if op_test_def.kind.arity == 2:
            input_vecs_with_arg1_nones = []
            input_vecs_with_arg1_nones.append(input_vecs[0] * 2)
            input_vecs_with_arg1_nones.append(
                input_vecs[1] + [None] * len(input_vecs[1])
            )
            variants.append(
                base_variant.variant(
                    "arg1-none",
                    input_vecs=tuple(input_vecs_with_arg1_nones),
                    expected_object_type=weave.types.optional(
                        base_variant.expected_object_type
                    ),
                    expected=base_variant.expected + [None] * len(input_vecs[1]),
                )
            )

    return variants


def make_binary_vectorscalar_variants(
    op_test_def: OpSpec,
    vector_position: typing.Union[typing.Literal["lhs"], typing.Literal["rhs"]],
) -> list[OpTestCaseBinaryVectorScalar]:
    op_name = op_test_def.op.name
    if vector_position == "lhs":
        vec_index = 0
        scalar_index = 1
    else:
        vec_index = 1
        scalar_index = 0

    variants: list[OpTestCaseBinaryVectorScalar] = []

    key_fn = lambda x: str(x.input[scalar_index])
    scalar_grouped = itertools.groupby(
        sorted(op_test_def.test_cases, key=key_fn), key=key_fn
    )
    scalar_groups: list[list[OpSpecTestCase]] = []
    for k, v in scalar_grouped:
        scalar_groups.append(list(v))
    groups_of_more_than_one = [g for g in scalar_groups if len(g) > 1]
    assert (
        len(groups_of_more_than_one) > 0
    ), f"You must provide at least two test cases with the same {vector_position.upper()} for op: {op_name}"

    for i, tcs in enumerate(groups_of_more_than_one):
        vec = [tc.input[vec_index] for tc in tcs]
        scalar = tcs[0].input[scalar_index]
        expected_object_type = weave.types.union(*[tc.expected_type for tc in tcs])
        result = [tc.expected for tc in tcs]
        vecgroup_variants: list[OpTestCaseBinaryVectorScalar] = []

        base_vec = OpTestCaseBinaryVectorScalar(
            [op_name, f"vector-{vector_position}", f"casegroup{i}"],
            op_test_def.op,
            vec=vec,
            scalar=scalar,
            vec_first=vector_position == "lhs",
            expected_object_type=expected_object_type,
            expected=result,
        )
        vecgroup_variants.append(base_vec)

        is_null_consuming = not language_nullability.should_force_none_result(
            {"first_arg": None}, op_test_def.op
        )

        if not is_null_consuming:
            vecgroup_variants.append(
                base_vec.variant(
                    "vec-with-none",
                    vec=vec + [None],
                    expected_object_type=weave.types.optional(expected_object_type),
                    expected=result + [None],
                )
            )

            if vector_position == "lhs":
                # We can't use the variant when vector_position is rhs, because
                # nullability means we'll just return None for the None scalar, not
                # using our vectorized op.
                # TODO: fix! We should return a vector of None in this case.
                vecgroup_variants.append(
                    base_vec.variant(
                        "scalar-none",
                        scalar=None,
                        expected_object_type=weave.types.optional(expected_object_type),
                        expected=[None] * len(vec),
                    )
                )

        variants.extend(vecgroup_variants)

    return variants


def is_all_none(x):
    if isinstance(x, list):
        return all(is_all_none(i) for i in x)
    elif isinstance(x, TaggedValue):
        return is_all_none(x.value)
    else:
        return x == None


def make_op_variants(op_test_def: OpSpec):
    variants: list[OpTestCase] = []
    op_name = op_test_def.op.name

    # standard, non-vectorized variants
    for i, test_case in enumerate(op_test_def.test_cases):
        base_variants = make_standard_variants(op_test_def, test_case, i)

        variants.extend(base_variants)

        # Add vectorized dim0 variants of all base_variants for list ops
        if op_test_def.kind.list_op:
            for variant in base_variants:
                # Don't need to test when first arg is None
                if not is_all_none(variant.input[0]):
                    variants.append(variant.vecdim0_variant("vecdim0"))
        if op_test_def.kind.list_list_op:
            for variant in base_variants:
                # Don't need to test when first arg is None
                if not is_all_none(variant.input[0]):
                    variants.append(variant.vecdim1_variant("vecdim1"))

    # All test cases in one vectorized call
    if op_test_def.kind.arity <= 2:
        variants.extend(make_standard_vector_variants(op_test_def))

    # (a, a) form where a is a vector
    if op_test_def.kind.arity == 2 and op_test_def.kind.uniform_input_types:
        all_lhs = [tc.input[0] for tc in op_test_def.test_cases]
        vecvec = OpTestCaseBinaryVectorVector(
            [op_name, "vecvec"], op_test_def.op, op_test_def.kind.arity, all_lhs
        )
        variants.append(vecvec)

    # (vec, scalar) and (scalar, vec) variants
    if op_test_def.kind.arity == 2:
        vec_variants = make_binary_vectorscalar_variants(
            op_test_def, vector_position="lhs"
        )
        if op_test_def.kind.uniform_input_types:
            if op_test_def.kind.commutative:
                for vec_variant in list(vec_variants):
                    # vector_rhs-none_commutative does not currently work because of the way we do nullability!
                    # We have op(None, [vec]) in this case, and we want a vector of nulls, but nullability
                    # returns None immediately at a very early stage.
                    # TODO: fix!
                    if vec_variant.scalar != None:
                        variants.append(
                            vec_variant.variant("commutative", vec_first=False)
                        )
            else:
                vec_variants.extend(
                    make_binary_vectorscalar_variants(
                        op_test_def, vector_position="rhs"
                    )
                )
        variants.extend(vec_variants)

    return variants


def make_all_variants() -> list[tuple[str, OpTestCase]]:
    variants: list[OpTestCase] = []
    for op_test_def in OP_TEST_SPECS:
        variants.extend(make_op_variants(op_test_def))
    return [("_".join(v.name), v) for v in variants]


VARIANTS = dict(make_all_variants())

SKIPS = set(
    [
        # test doesn't produce correct expected output
        "joinAll_case1_vecdim1",
        # op implementation produces incorrect output
        "joinAll_case0_arg0-tag_vecdim1",
        "joinAll_case1_arg0-tag_vecdim1",
    ]
)


@pytest.mark.parametrize("name", [name for name in VARIANTS.keys()])
def test_all_ops(name: str):
    if name in SKIPS:
        return
    test = VARIANTS[name]
    test.check()



from typing import Generator

import pytest

import weave
from weave.integrations.integration_utilities import (
    filter_body,
    flatten_calls,
    flattened_calls_to_names,
)
from weave.trace.weave_client import Call
from weave.trace_server.trace_server_interface import CallsFilter


def assert_calls_correct_for_quickstart(flattened_calls: list[Call]) -> None:
    """Next, the major thing to assert is the "shape" of the calls:
    llama_index.query
        llama_index.retrieve
            llama_index.embedding
        llama_index.synthesize
            llama_index.chunking
            llama_index.chunking
            llama_index.templating
            llama_index.llm
                openai.chat.completions.create
    """
    assert len(flattened_calls) == 9

    exp = [
        ("llama_index.query", 0),
        ("llama_index.retrieve", 1),
        ("llama_index.embedding", 2),
        ("llama_index.synthesize", 1),
        ("llama_index.chunking", 2),
        ("llama_index.chunking", 2),
        ("llama_index.templating", 2),
        ("llama_index.llm", 2),
        ("openai.chat.completions.create", 3),
    ]
    assert flattened_calls_to_names(flattened_calls) == exp


@pytest.fixture
def fake_api_key() -> Generator[None, None, None]:
    import os

    orig_key = os.environ.get("OPENAI_API_KEY")
    os.environ["OPENAI_API_KEY"] = "sk-DUMMY_KEY"
    try:
        yield
    finally:
        if orig_key is None:
            del os.environ["OPENAI_API_KEY"]
        else:
            os.environ["OPENAI_API_KEY"] = orig_key


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
def test_llamaindex_quickstart(
    client: weave.trace.weave_client.WeaveClient, fake_api_key: None
) -> None:
    # This is taken directly from  https://docs.llamaindex.ai/en/stable/getting_started/starter_example/
    from llama_index.core import SimpleDirectoryReader, VectorStoreIndex

    documents = SimpleDirectoryReader("integrations/llamaindex/test_data").load_data()
    index = VectorStoreIndex.from_documents(documents)

    query_engine = index.as_query_engine()
    response = query_engine.query("What did the author do growing up?")

    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert_calls_correct_for_quickstart(flattened_calls)
    call, _ = flattened_calls[-2]
    assert call.inputs["serialized"]["api_key"] == "REDACTED"


@pytest.mark.skip_clickhouse_client  # TODO:VCR recording does not seem to allow us to make requests to the clickhouse db in non-recording mode
@pytest.mark.vcr(
    filter_headers=["authorization"],
    allowed_hosts=["api.wandb.ai", "localhost", "trace.wandb.ai"],
    before_record_request=filter_body,
)
@pytest.mark.asyncio
async def test_llamaindex_quickstart_async(
    client: weave.trace.weave_client.WeaveClient, fake_api_key: None
) -> None:
    # This is taken directly from  https://docs.llamaindex.ai/en/stable/getting_started/starter_example/
    from llama_index.core import SimpleDirectoryReader, VectorStoreIndex

    documents = SimpleDirectoryReader("integrations/llamaindex/test_data").load_data()
    index = VectorStoreIndex.from_documents(documents)

    query_engine = index.as_query_engine()
    response = await query_engine.aquery("What did the author do growing up?")

    calls = list(client.calls(filter=CallsFilter(trace_roots_only=True)))
    flattened_calls = flatten_calls(calls)
    assert_calls_correct_for_quickstart(flattened_calls)
    call, _ = flattened_calls[-2]
    assert call.inputs["serialized"]["api_key"] == "REDACTED"
